#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import unittest

from servicecatalog_puppet import constants, task_reference_constants


class SSMParameterHandlerTest(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None
        from servicecatalog_puppet.commands.task_reference_helpers.generators import (
            ssm_parameter_handler,
        )

        self.sut = ssm_parameter_handler

    def test_for_ssm_parameters_for_spoke_execution_mode(self):
        # setup
        puppet_account_id = "hub_account_id"
        account_id = "spoke_account_id"
        region = "eu-west-0"
        default_region = "eu-south-0"
        item_name = "depsrefactor"
        dependency_item_name = "something_else"

        task = {
            "stack_name": item_name,
            "launch_name": "",
            "stack_set_name": "",
            "capabilities": ["CAPABILITY_NAMED_IAM"],
            "ssm_param_outputs": [],
            "bucket": f"sc-puppet-stacks-repository-{puppet_account_id}",
            "key": f"stack/{item_name}/v1/stack.template-${{AWS::Region}}.yaml",
            "version_id": "",
            "execution": "spoke",
            "use_service_role": True,
            "tags": [],
            "puppet_account_id": puppet_account_id,
            "status": None,
            "requested_priority": 0,
            "dependencies": [
                {
                    "affinity": "stack",
                    "name": dependency_item_name,
                    "type": "stack",
                }
            ],
            "account_id": account_id,
            "region": region,
            task_reference_constants.MANIFEST_SECTION_NAMES: {"stacks": True},
            task_reference_constants.MANIFEST_ITEM_NAMES: {item_name: True},
            task_reference_constants.MANIFEST_ACCOUNT_IDS: {account_id: True},
            "section_name": "stacks",
            "item_name": item_name,
            "dependencies_by_reference": [
                "create-policies",
                f"prepare-account-for-stacks-{account_id}",
                f"get-template-from-s3-stacks-{item_name}",
            ],
            "task_reference": f"{account_id}-{region}",
            "get_s3_template_ref": f"get-template-from-s3-stacks-{item_name}",
        }
        task_reference = task["task_reference"]
        all_tasks = {task_reference: task}

        parameter_path = "/sleeper/SleeperFunctionArn/${AWS::Region}/stacks"

        parameter_details = {
            "ssm": {"name": parameter_path, "account_id": "${AWS::AccountId}"}
        }

        # exercise
        self.sut.ssm_parameter_handler(
            all_tasks,
            default_region,
            all_tasks,
            parameter_details,
            puppet_account_id,
            task,
        )

        # verify
        n_all_tasks = len(all_tasks.keys())
        self.assertEqual(
            [
                "create-policies",
                f"prepare-account-for-stacks-{account_id}",
                f"get-template-from-s3-stacks-{item_name}",
                f"{constants.SSM_PARAMETERS}-{account_id}-{default_region}-{f'/sleeper/SleeperFunctionArn/{region}/stacks'}",
            ],
            all_tasks[task_reference].get("dependencies_by_reference"),
            "assert new dependency is added to the task needing the parameter",
        )

        expected_ssm_parameter_task_ref = f"{constants.SSM_PARAMETERS}-{account_id}-{default_region}-{f'/sleeper/SleeperFunctionArn/{region}/stacks'}"
        self.assertEqual(
            {
                "account_id": account_id,
                "region": default_region,
                task_reference_constants.MANIFEST_SECTION_NAMES: task.get(
                    task_reference_constants.MANIFEST_SECTION_NAMES
                ),
                task_reference_constants.MANIFEST_ITEM_NAMES: task.get(
                    task_reference_constants.MANIFEST_ITEM_NAMES
                ),
                task_reference_constants.MANIFEST_ACCOUNT_IDS: task.get(
                    task_reference_constants.MANIFEST_ACCOUNT_IDS
                ),
                "execution": task.get("execution"),
                "jmespath": None,
                "param_name": f"/sleeper/SleeperFunctionArn/{region}/stacks",
                "section_name": constants.SSM_PARAMETERS,
                "task_reference": expected_ssm_parameter_task_ref,
                "dependencies_by_reference": [],
                "dependencies": [],
            },
            all_tasks[expected_ssm_parameter_task_ref],
            "assert the boto3 task is generated correctly",
        )

        self.assertEqual(2, n_all_tasks, "assert the correct number of tasks exist")

    def test_for_ssm_parameters_for_hub_execution_mode(self):
        # setup
        puppet_account_id = "hub_account_id"
        account_id = "spoke_account_id"
        region = "eu-west-0"
        default_region = "eu-south-0"
        item_name = "depsrefactor"
        dependency_item_name = "something_else"

        task = {
            "stack_name": item_name,
            "launch_name": "",
            "stack_set_name": "",
            "capabilities": ["CAPABILITY_NAMED_IAM"],
            "ssm_param_outputs": [],
            "bucket": f"sc-puppet-stacks-repository-{puppet_account_id}",
            "key": f"stack/{item_name}/v1/stack.template-${{AWS::Region}}.yaml",
            "version_id": "",
            "execution": "hub",
            "use_service_role": True,
            "tags": [],
            "puppet_account_id": puppet_account_id,
            "status": None,
            "requested_priority": 0,
            "dependencies": [
                {
                    "affinity": "stack",
                    "name": dependency_item_name,
                    "type": "stack",
                }
            ],
            "account_id": account_id,
            "region": region,
            task_reference_constants.MANIFEST_SECTION_NAMES: {"stacks": True},
            task_reference_constants.MANIFEST_ITEM_NAMES: {item_name: True},
            task_reference_constants.MANIFEST_ACCOUNT_IDS: {account_id: True},
            "section_name": "stacks",
            "item_name": item_name,
            "dependencies_by_reference": [
                "create-policies",
                f"prepare-account-for-stacks-{account_id}",
                f"get-template-from-s3-stacks-{item_name}",
            ],
            "task_reference": f"{account_id}-{region}",
            "get_s3_template_ref": f"get-template-from-s3-stacks-{item_name}",
        }
        task_reference = task["task_reference"]
        all_tasks = {task_reference: task}

        parameter_path = "/sleeper/SleeperFunctionArn/${AWS::Region}/stacks"

        parameter_details = {
            "ssm": {"name": parameter_path, "account_id": "${AWS::AccountId}"}
        }

        # exercise
        self.sut.ssm_parameter_handler(
            all_tasks,
            default_region,
            all_tasks,
            parameter_details,
            puppet_account_id,
            task,
        )

        # verify
        n_all_tasks = len(all_tasks.keys())
        self.assertEqual(
            [
                "create-policies",
                f"prepare-account-for-stacks-{account_id}",
                f"get-template-from-s3-stacks-{item_name}",
                f"{constants.SSM_PARAMETERS}-{account_id}-{default_region}-{f'/sleeper/SleeperFunctionArn/{region}/stacks'}",
            ],
            all_tasks[task_reference].get("dependencies_by_reference"),
            "assert new dependency is added to the task needing the parameter",
        )

        expected_ssm_parameter_task_ref = f"{constants.SSM_PARAMETERS}-{account_id}-{default_region}-{f'/sleeper/SleeperFunctionArn/{region}/stacks'}"
        self.assertEqual(
            {
                "account_id": account_id,
                "region": default_region,
                task_reference_constants.MANIFEST_SECTION_NAMES: task.get(
                    task_reference_constants.MANIFEST_SECTION_NAMES
                ),
                task_reference_constants.MANIFEST_ITEM_NAMES: task.get(
                    task_reference_constants.MANIFEST_ITEM_NAMES
                ),
                task_reference_constants.MANIFEST_ACCOUNT_IDS: task.get(
                    task_reference_constants.MANIFEST_ACCOUNT_IDS
                ),
                "execution": task.get("execution"),
                "jmespath": None,
                "param_name": f"/sleeper/SleeperFunctionArn/{region}/stacks",
                "section_name": constants.SSM_PARAMETERS,
                "task_reference": expected_ssm_parameter_task_ref,
                "dependencies_by_reference": [],
                "dependencies": [],
            },
            all_tasks[expected_ssm_parameter_task_ref],
            "assert the boto3 task is generated correctly",
        )

        self.assertEqual(2, n_all_tasks, "assert the correct number of tasks exist")

    def test_for_ssm_parameters_for_spoke_execution_mode_where_param_is_set_in_hub(
        self,
    ):
        # setup
        puppet_account_id = "hub_account_id"
        account_id = "spoke_account_id"
        region = "eu-west-0"
        default_region = "eu-south-0"
        item_name = "depsrefactor"
        dependency_item_name = "something_else"

        dependency_task = {
            "stack_name": dependency_item_name,
            "launch_name": "",
            "stack_set_name": "",
            "capabilities": ["CAPABILITY_NAMED_IAM"],
            "ssm_param_outputs": [],
            "bucket": f"sc-puppet-stacks-repository-{puppet_account_id}",
            "key": f"stack/{dependency_item_name}/v1/stack.template-${{AWS::Region}}.yaml",
            "version_id": "",
            "execution": "hub",
            "use_service_role": True,
            "tags": [],
            "puppet_account_id": puppet_account_id,
            "status": None,
            "requested_priority": 0,
            "dependencies": [],
            "account_id": puppet_account_id,
            "region": region,
            task_reference_constants.MANIFEST_SECTION_NAMES: {"stacks": True},
            task_reference_constants.MANIFEST_ITEM_NAMES: {dependency_item_name: True},
            task_reference_constants.MANIFEST_ACCOUNT_IDS: {account_id: True},
            "section_name": "stacks",
            "item_name": dependency_item_name,
            "dependencies_by_reference": [
                "create-policies",
                f"prepare-account-for-stacks-{account_id}",
                f"get-template-from-s3-stacks-{dependency_item_name}",
            ],
            "task_reference": f"{dependency_item_name}-{account_id}-{region}",
            "get_s3_template_ref": f"get-template-from-s3-stacks-{dependency_item_name}",
        }
        task = {
            "stack_name": item_name,
            "launch_name": "",
            "stack_set_name": "",
            "capabilities": ["CAPABILITY_NAMED_IAM"],
            "ssm_param_outputs": [],
            "bucket": f"sc-puppet-stacks-repository-{puppet_account_id}",
            "key": f"stack/{item_name}/v1/stack.template-${{AWS::Region}}.yaml",
            "version_id": "",
            "execution": "spoke",
            "use_service_role": True,
            "tags": [],
            "puppet_account_id": puppet_account_id,
            "status": None,
            "requested_priority": 0,
            "dependencies": [
                {
                    "affinity": "stack",
                    "name": dependency_item_name,
                    "type": "stack",
                }
            ],
            "account_id": account_id,
            "region": region,
            task_reference_constants.MANIFEST_SECTION_NAMES: {"stacks": True},
            task_reference_constants.MANIFEST_ITEM_NAMES: {item_name: True},
            task_reference_constants.MANIFEST_ACCOUNT_IDS: {account_id: True},
            "section_name": "stacks",
            "item_name": item_name,
            "dependencies_by_reference": [
                "create-policies",
                f"prepare-account-for-stacks-{account_id}",
                f"get-template-from-s3-stacks-{item_name}",
            ],
            "task_reference": f"{account_id}-{region}",
            "get_s3_template_ref": f"get-template-from-s3-stacks-{item_name}",
        }
        task_reference = task["task_reference"]
        all_tasks = {task_reference: task}

        parameter_path = "/sleeper/SleeperFunctionArn/${AWS::Region}/stacks"

        parameter_details = {
            "ssm": {
                "name": parameter_path,
            }
        }

        # exercise
        self.sut.ssm_parameter_handler(
            all_tasks,
            default_region,
            all_tasks,
            parameter_details,
            puppet_account_id,
            task,
        )

        # verify
        n_all_tasks = len(all_tasks.keys())
        self.assertEqual(
            [
                "create-policies",
                f"prepare-account-for-stacks-{account_id}",
                f"get-template-from-s3-stacks-{item_name}",
                f"{constants.SSM_PARAMETERS}-{puppet_account_id}-{default_region}-{f'/sleeper/SleeperFunctionArn/{region}/stacks'}",
            ],
            all_tasks[task_reference].get("dependencies_by_reference"),
            "assert new dependency is added to the task needing the parameter",
        )

        expected_ssm_parameter_task_ref = f"{constants.SSM_PARAMETERS}-{puppet_account_id}-{default_region}-{f'/sleeper/SleeperFunctionArn/{region}/stacks'}"
        self.assertEqual(
            {
                "account_id": puppet_account_id,
                "region": default_region,
                task_reference_constants.MANIFEST_SECTION_NAMES: task.get(
                    task_reference_constants.MANIFEST_SECTION_NAMES
                ),
                task_reference_constants.MANIFEST_ITEM_NAMES: task.get(
                    task_reference_constants.MANIFEST_ITEM_NAMES
                ),
                task_reference_constants.MANIFEST_ACCOUNT_IDS: task.get(
                    task_reference_constants.MANIFEST_ACCOUNT_IDS
                ),
                "execution": dependency_task.get("execution"),
                "jmespath": None,
                "param_name": f"/sleeper/SleeperFunctionArn/{region}/stacks",
                "section_name": constants.SSM_PARAMETERS,
                "task_reference": expected_ssm_parameter_task_ref,
                "dependencies_by_reference": [],
                "dependencies": [],
            },
            all_tasks[expected_ssm_parameter_task_ref],
            "assert the boto3 task is generated correctly",
        )

        self.assertEqual(2, n_all_tasks, "assert the correct number of tasks exist")



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import time

import luigi

from servicecatalog_puppet import constants
from servicecatalog_puppet.workflow.dependencies import tasks


class CopyIntoSpokeLocalPortfolioTask(tasks.TaskWithReference):
    account_id = luigi.Parameter()
    region = luigi.Parameter()
    portfolio_task_reference = luigi.Parameter()

    portfolio_get_all_products_and_their_versions_ref = luigi.Parameter()
    portfolio_get_all_products_and_their_versions_for_hub_ref = luigi.Parameter()
    cachable_level = constants.CACHE_LEVEL_RUN

    def params_for_results_display(self):
        return {
            "task_reference": self.task_reference,
            "puppet_account_id": self.puppet_account_id,
            "region": self.region,
            "account_id": self.account_id,
        }

    def run(self):
        spoke_portfolio_details = self.get_output_from_reference_dependency(
            self.portfolio_task_reference
        )
        spoke_portfolio_id = spoke_portfolio_details.get("Id")
        spoke_products_and_their_versions = self.get_output_from_reference_dependency(
            self.portfolio_get_all_products_and_their_versions_ref
        )
        hub_products_and_their_versions = self.get_output_from_reference_dependency(
            self.portfolio_get_all_products_and_their_versions_for_hub_ref
        )

        copy_product_tokens = list()
        versions_requiring_updates = dict()
        products_requiring_adding_to_portfolio = dict()
        with self.spoke_regional_client("servicecatalog") as servicecatalog:
            for (
                hub_product_name,
                hub_product_details,
            ) in hub_products_and_their_versions.items():
                versions_to_copy = list()
                args_to_use = dict(
                    SourceProductArn=hub_product_details.get("ProductArn"),
                    SourceProvisioningArtifactIdentifiers=versions_to_copy,
                    CopyOptions=[
                        "CopyTags",
                    ],
                )
                hub_versions_details = hub_product_details.get("Versions", {})
                if spoke_products_and_their_versions.get(hub_product_name):
                    args_to_use["TargetProductId"] = (
                        spoke_products_and_their_versions.get(hub_product_name).get(
                            "ProductId"
                        )
                    )
                else:
                    products_requiring_adding_to_portfolio[hub_product_name] = True

                spoke_product_details = spoke_products_and_their_versions.get(
                    hub_product_name, {}
                )
                spoke_versions_details = spoke_product_details.get("Versions", {})
                version_names_to_ignore = ["-"] + list(spoke_versions_details.keys())
                for (
                    hub_version_name,
                    hub_version_details,
                ) in hub_versions_details.items():
                    if hub_version_name not in version_names_to_ignore:
                        versions_to_copy.append(
                            dict(
                                Id=hub_version_details.get("Id"),
                            )
                        )
                    else:
                        if hub_version_name == "-":
                            continue
                        spoke_product_id = spoke_product_details["ProductId"]
                        if not versions_requiring_updates.get(spoke_product_id):
                            versions_requiring_updates[spoke_product_id] = dict()
                        spoke_version_id = spoke_versions_details[hub_version_name][
                            "Id"
                        ]

                        versions_requiring_updates[spoke_product_id][
                            spoke_version_id
                        ] = dict(
                            Active=hub_version_details.get("Active"),
                            Guidance=hub_version_details.get("Guidance"),
                            Description=hub_version_details.get("Description"),
                        )

                if len(versions_to_copy) > 0:
                    copy_product_tokens.append(
                        (
                            hub_product_name,
                            servicecatalog.copy_product(**args_to_use).get(
                                "CopyProductToken"
                            ),
                        )
                    )
            self.info("Finished copying products")

            while len(copy_product_tokens) > 0:
                first_item_in_list = copy_product_tokens[0]
                product_name, copy_product_token_to_check = first_item_in_list
                response = servicecatalog.describe_copy_product_status(
                    CopyProductToken=copy_product_token_to_check
                )
                copy_product_status = response.get("CopyProductStatus")
                if copy_product_status == "SUCCEEDED":
                    if products_requiring_adding_to_portfolio.get(product_name):
                        products_requiring_adding_to_portfolio[product_name] = (
                            response.get("TargetProductId")
                        )
                    copy_product_tokens.remove(first_item_in_list)
                elif copy_product_status == "FAILED":
                    raise Exception(f"Failed to copy product {copy_product_status}")
                elif copy_product_status == "IN_PROGRESS":
                    time.sleep(1)
                else:
                    raise Exception(f"Not handled copy product status {response}")
        self.info("Finished waiting for copy products")

        for product_name, product_id in products_requiring_adding_to_portfolio.items():
            servicecatalog.associate_product_with_portfolio(
                ProductId=product_id,
                PortfolioId=spoke_portfolio_id,
            )
        self.info("Finished associating products")

        throttle = 0
        for product_id, product_details in versions_requiring_updates.items():
            for version_id, version_details in product_details.items():
                servicecatalog.update_provisioning_artifact(
                    ProductId=product_id,
                    ProvisioningArtifactId=version_id,
                    **version_details,
                )
                throttle += 1
                if throttle == 4:
                    time.sleep(1.1)
                    throttle = 0
        self.info("Finished updating versions that were copied")

        products_to_check = list(products_requiring_adding_to_portfolio.values())
        n_products_to_check = len(products_to_check)
        products_found = 0
        while products_found < n_products_to_check:
            products_ids = list()
            paginator = servicecatalog.get_paginator("search_products_as_admin")
            for page in paginator.paginate(
                PortfolioId=spoke_portfolio_id,
            ):
                for pvd in page.get("ProductViewDetails", []):
                    products_ids.append(pvd.get("ProductViewSummary").get("ProductId"))
            products_found = 0
            for product_to_check in products_to_check:
                if product_to_check in products_ids:
                    products_found += 1
        self.info("Finished waiting for association of products to portfolio")
        self.write_empty_output()



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import luigi
import yaml

import servicecatalog_puppet.manifest_utils
import servicecatalog_puppet.serialisation_utils
from servicecatalog_puppet import constants
from servicecatalog_puppet.workflow.dependencies import tasks


class DoExecuteSimulatePolicyTask(tasks.TaskWithReference):

    simulate_policy_name = luigi.Parameter()
    puppet_account_id = luigi.Parameter()

    region = luigi.Parameter()
    account_id = luigi.Parameter()

    execution = luigi.Parameter()

    requested_priority = luigi.IntParameter()

    simulation_type = luigi.Parameter()
    policy_source_arn = luigi.Parameter()
    policy_input_list = luigi.ListParameter()
    permissions_boundary_policy_input_list = luigi.ListParameter()
    action_names = luigi.ListParameter()
    expected_decision = luigi.Parameter()
    resource_arns = luigi.ListParameter()
    resource_policy = luigi.Parameter()
    resource_owner = luigi.Parameter()
    caller_arn = luigi.Parameter()
    context_entries = luigi.ListParameter()
    resource_handling_option = luigi.Parameter()
    cachable_level = constants.CACHE_LEVEL_RUN

    def params_for_results_display(self):
        return {
            "puppet_account_id": self.puppet_account_id,
            "simulate_policy_name": self.simulate_policy_name,
            "region": self.region,
            "account_id": self.account_id,
        }

    def run(self):
        with self.spoke_regional_client("iam") as iam:
            kwargs = dict(ActionNames=self.action_names)
            if len(self.policy_input_list) > 0:
                kwargs["PolicyInputList"] = self.policy_input_list

            if len(self.permissions_boundary_policy_input_list) > 0:
                kwargs["PermissionsBoundaryPolicyInputList"] = (
                    self.permissions_boundary_policy_input_list
                )

            if len(self.resource_arns) > 0:
                kwargs["ResourceArns"] = self.resource_arns

            if self.resource_policy != "":
                kwargs["ResourcePolicy"] = self.resource_policy

            if self.resource_owner != "":
                kwargs["ResourceOwner"] = self.resource_owner

            if self.caller_arn != "":
                kwargs["CallerArn"] = self.caller_arn

            if len(self.context_entries) > 0:
                kwargs["ContextEntries"] = (
                    servicecatalog_puppet.serialisation_utils.unwrap(
                        self.context_entries
                    )
                )

            if self.resource_handling_option != "":
                kwargs["ResourceHandlingOption"] = self.resource_handling_option

            if self.simulation_type == "principal":
                kwargs["PolicySourceArn"] = self.policy_source_arn.replace(
                    "${AWS::AccountId}", self.account_id
                )
                result = iam.simulate_principal_policy(**kwargs)
            else:
                if len(self.policy_input_list) == 0:
                    raise Exception(
                        "policy_input_list is required when simulation_type is 'custom'"
                    )
                result = iam.simulate_custom_policy(**kwargs)

            failures = list()
            for evaluation_result in result.get("EvaluationResults"):
                if evaluation_result.get("EvalDecision") != self.expected_decision:
                    failures.append(evaluation_result)

            if len(failures) > 0:
                raise Exception(
                    f"{len(failures)} unexpected decision(s) encountered:\n{yaml.safe_dump(failures)}"
                )

        self.write_output(result)



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

from unittest import skip

from servicecatalog_puppet.workflow import tasks_unit_tests_helper


class GetSSMParameterTaskTest(tasks_unit_tests_helper.PuppetTaskUnitTest):
    account_id = "account_id"
    param_name = "param_name"
    region = "region"
    jmespath_location = None

    def setUp(self) -> None:
        from servicecatalog_puppet.workflow.ssm import get_ssm_parameter_task

        self.module = get_ssm_parameter_task

        self.sut = self.module.GetSSMParameterTask(
            **self.get_common_args(),
            account_id=self.account_id,
            param_name=self.param_name,
            region=self.region,
            jmespath_location=self.jmespath_location,
        )

        self.wire_up_mocks()

    def test_params_for_results_display(self):
        # setup
        expected_result = {
            "task_reference": self.task_reference,
            "account_id": self.account_id,
            "region": self.region,
            "param_name": self.param_name,
            "jmespath_location": self.jmespath_location,
        }

        # exercise
        actual_result = self.sut.params_for_results_display()

        # verify
        self.assertEqual(expected_result, actual_result)

    def test_run(self):
        # setup
        parameter = dict(
            this="that",
        )
        expected_output = {self.param_name: parameter}
        self.add_expected_request_and_response(
            tasks_unit_tests_helper.SPOKE_REGIONAL_CLIENT,
            "ssm",
            "get_parameter",
            dict(Name=self.sut.get_parameter_name_to_use()),
            dict(Parameter=parameter),
        )

        # exercise
        self.sut.run()

        # verify
        self.assert_output(expected_output)


class GetSSMParameterByPathTaskTest(tasks_unit_tests_helper.PuppetTaskUnitTest):
    account_id = "account_id"
    path = "path"
    region = "region"
    jmespath_location = None

    def setUp(self) -> None:
        from servicecatalog_puppet.workflow.ssm import get_ssm_parameter_task

        self.module = get_ssm_parameter_task

        self.sut = self.module.GetSSMParameterByPathTask(
            **self.get_common_args(),
            account_id=self.account_id,
            path=self.path,
            region=self.region,
            jmespath_location=self.jmespath_location,
        )

        self.wire_up_mocks()

    def test_params_for_results_display(self):
        # setup
        expected_result = {
            "task_reference": self.task_reference,
            "account_id": self.account_id,
            "region": self.region,
            "path": self.path,
            "jmespath_location": self.jmespath_location,
        }

        # exercise
        actual_result = self.sut.params_for_results_display()

        # verify
        self.assertEqual(expected_result, actual_result)

    def test_run(self):
        parameter = dict(Name="that", Value="hello")
        expected_output = dict(that=parameter)
        self.add_expected_paginated_request_and_response(
            tasks_unit_tests_helper.SPOKE_REGIONAL_CLIENT,
            "ssm",
            "get_parameters_by_path",
            dict(Path=self.sut.get_parameter_path_to_use(), Recursive=True),
            [
                dict(Parameters=[parameter]),
            ],
        )

        # exercise
        self.sut.run()

        # verify
        self.assert_output(expected_output)



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import os

import pkg_resources


PREFIX = "SC-P--"
BOOTSTRAP_STACK_NAME = "servicecatalog-puppet"
PIPELINE_NAME = "servicecatalog-puppet-pipeline"
SERVICE_CATALOG_PUPPET_REPO_NAME = "ServiceCatalogPuppet"
OUTPUT = "output"
TEMPLATES = os.path.sep.join([OUTPUT, "templates"])
LAUNCHES_PATH = os.path.sep.join([OUTPUT, "launches"])
CONFIG_PARAM_NAME = "/servicecatalog-puppet/config"
CONFIG_PARAM_NAME_ORG_IAM_ROLE_ARN = "/servicecatalog-puppet/org-iam-role-arn"
CONFIG_PARAM_NAME_ORG_SCP_ROLE_ARN = "/servicecatalog-puppet/org-scp-role-arn"
PUPPET_ORG_ROLE_FOR_EXPANDS_ARN = "PuppetOrgRoleForExpandsArn"
HOME_REGION_PARAM_NAME = "/servicecatalog-puppet/home-region"

PROVISIONED = "provisioned"
TERMINATED = "terminated"

DEFAULT_TIMEOUT = 0
PARAMETERS = "parameters"
ACCOUNTS = "accounts"
MAPPINGS = "mappings"
LAUNCHES = "launches"
STACK = "stack"
STACKS = "stacks"
SPOKE_LOCAL_PORTFOLIOS = "spoke-local-portfolios"
IMPORTED_PORTFOLIOS = "imported-portfolios"
LAMBDA_INVOCATIONS = "lambda-invocations"
ASSERTIONS = "assertions"
WORKSPACES = "workspaces"
CFCT = "cfct"
WORKSPACE = "workspace"
APPS = "apps"
APP = "app"
ASSERTION = "assertion"
CODE_BUILD_RUNS = "code-build-runs"
CODE_BUILD_RUN = "code-build-run"
SERVICE_CONTROL_POLICIES = "service-control-policies"
SERVICE_CONTROL_POLICY = "service-control-policy"
TAG_POLICIES = "tag-policies"
TAG_POLICY = "tag-policy"
C7N_AWS_CLOUDTRAILS = "c7n-aws-cloudtrails"
C7N_AWS_CLOUDTRAIL = "c7n-aws-cloudtrail"
C7N_AWS_LAMBDAS = "c7n-aws-lambdas"
C7N_AWS_LAMBDA = "c7n-aws-lambda"

ORGANIZATIONAL_UNIT = "organizational-unit"
ORGANIZATIONAL_UNITS = "organizational-units"

SIMULATE_POLICIES = "simulate-policies"
SIMULATE_POLICY = "simulate-policy"

ACTIONS = "actions"

LAUNCH = "launch"
LAMBDA_INVOCATION = "lambda-invocation"
SPOKE_LOCAL_PORTFOLIO = "spoke-local-portfolio"
IMPORTED_PORTFOLIO = "imported-portfolio"
ACTION = "action"

AFFINITY = "affinity"
AFFINITY_REGION = "region"
AFFINITY_ACCOUNT = "account"
AFFINITY_ACCOUNT_AND_REGION = "account-and-region"

RESULTS_DIRECTORY = "results"

NO_CHANGE = "NO_CHANGE"
CHANGE = "CHANGE"

EVENT_BUS_NAME = "servicecatalog-puppet-event-bus"
EVENT_BUS_IN_SPOKE_NAME = "servicecatalog-puppet-spoke-event-bus"
SERVICE_CATALOG_PUPPET_EVENT_SOURCE = "servicecatalog-puppet"
SERVICE_CATALOG_PUPPET_OPS_CENTER_SOURCE = "servicecatalog-puppet"

HOME_REGION = os.environ.get(
    "AWS_REGION", os.environ.get("AWS_DEFAULT_REGION", "eu-west-1")
)

EVENTBRIDGE_MAX_EVENTS_PER_CALL = 10

SPOKE_VERSION_SSM_PARAM_NAME = "service-catalog-puppet-spoke-version"
PUPPET_VERSION_SSM_PARAM_NAME = "service-catalog-puppet-version"
PUPPET_VERSION_INITIAL_SSM_PARAM_NAME = "service-catalog-puppet-version-initial"

SPOKE_LOCAL_PORTFOLIO_STATUS_SHARED = "shared"
SPOKE_LOCAL_PORTFOLIO_STATUS_TERMINATED = "terminated"

EXECUTION_MODE_ASYNC = "async"
EXECUTION_MODE_HUB = "hub"
EXECUTION_MODE_SPOKE = "spoke"
EXECUTION_MODE_HUB_AND_SPOKE_SPLIT = "hub-and-spoke-split"
EXECUTION_MODE_DEFAULT = EXECUTION_MODE_HUB
EXECUTION_SPOKE_CODEBUILD_PROJECT_NAME = "servicecatalog-puppet-deploy-in-spoke"

SHARING_MODE_ACCOUNT = "ACCOUNT"
SHARING_MODE_AWS_ORGANIZATIONS = "AWS_ORGANIZATIONS"
SHARING_MODE_DEFAULT = SHARING_MODE_ACCOUNT

SHARE_TAG_OPTIONS_DEFAULT = False

PARTITION_DEFAULT = "aws"
PARTITION_ENVIRONMENTAL_VARIABLE_NAME = "PARTITION"
PARTITION_SSM_PARAMETER_VARIABLE_NAME = "/servicecatalog-puppet/partition"

PUPPET_ROLE_NAME_DEFAULT = "PuppetRole"
PUPPET_STACK_ROLE_NAME_DEFAULT = "PuppetStackRole"
PUPPET_ROLE_NAME_ENVIRONMENTAL_VARIABLE_NAME = "PUPPET_ROLE_NAME"
PUPPET_STACK_ROLE_NAME_ENVIRONMENTAL_VARIABLE_NAME = "PUPPET_STACK_ROLE_NAME"
PUPPET_ROLE_NAME_SSM_PARAMETER_VARIABLE_NAME = "/servicecatalog-puppet/puppet-role/name"

PUPPET_ROLE_PATH_DEFAULT = "/servicecatalog-puppet/"
PUPPET_ROLE_PATH_ENVIRONMENTAL_VARIABLE_NAME = "PUPPET_ROLE_PATH"
PUPPET_ROLE_PATH_SSM_PARAMETER_VARIABLE_NAME = "/servicecatalog-puppet/puppet-role/path"

DEPLOY_ENVIRONMENT_COMPUTE_TYPE_DEFAULT = "BUILD_GENERAL1_SMALL"

START_SHARED_SCHEDULER_COMMAND = "luigid --background --pidfile luigi.pid --logdir results/logs --state-path results/state"

CONFIG_REGIONS = "regions"
CONFIG_SHOULD_COLLECT_CLOUDFORMATION_EVENTS = "should_collect_cloudformation_events"
CONFIG_SHOULD_USE_STACKS_SERVICE_ROLE = "should_use_stacks_service_role"
MANIFEST_SHOULD_USE_STACKS_SERVICE_ROLE = "use_stacks_service_role"
CONFIG_SHOULD_USE_STACKS_SERVICE_ROLE_DEFAULT = False
CONFIG_SHOULD_USE_SHARED_SCHEDULER = "should_use_shared_scheduler"
CONFIG_SHOULD_EXPLODE_MANIFEST = "should_explode_manifest"

PUBLISHED_VERSION = pkg_resources.require("aws-service-catalog-puppet")[0].version
VERSION_OVERRIDE = "SCP_VERSION_OVERRIDE"
VERSION = os.getenv(VERSION_OVERRIDE, PUBLISHED_VERSION)

PRODUCT_GENERATION_METHOD_IMPORT = "import"
PRODUCT_GENERATION_METHOD_COPY = "copy"
PRODUCT_GENERATION_METHOD_DEFAULT = PRODUCT_GENERATION_METHOD_COPY

LUIGI_DEFAULT_LOG_LEVEL = "INFO"

ALL_SECTION_NAMES = [
    LAUNCHES,
    STACKS,
    SPOKE_LOCAL_PORTFOLIOS,
    IMPORTED_PORTFOLIOS,
    LAMBDA_INVOCATIONS,
    CODE_BUILD_RUNS,
    ASSERTIONS,
    APPS,
    WORKSPACES,
    SERVICE_CONTROL_POLICIES,
    SIMULATE_POLICIES,
    TAG_POLICIES,
    C7N_AWS_LAMBDAS,
]
ALL_SECTION_NAMES_THAT_GENERATE_OUTPUTS = [
    LAUNCHES,
    STACKS,
    APPS,
    WORKSPACES,
]

ALL_SPOKE_EXECUTABLE_SECTION_NAMES = [
    LAUNCHES,
    STACKS,
    LAMBDA_INVOCATIONS,
    CODE_BUILD_RUNS,
    ASSERTIONS,
    APPS,
    WORKSPACES,
    SIMULATE_POLICIES,
]

ALL_SECTION_NAME_SINGULAR_AND_PLURAL_LIST = [
    (LAUNCH, LAUNCHES),
    (STACK, STACKS),
    (SPOKE_LOCAL_PORTFOLIO, SPOKE_LOCAL_PORTFOLIOS),
    (IMPORTED_PORTFOLIO, IMPORTED_PORTFOLIOS),
    (LAMBDA_INVOCATION, LAMBDA_INVOCATIONS),
    (CODE_BUILD_RUN, CODE_BUILD_RUNS),
    (ASSERTION, ASSERTIONS),
    (APP, APPS),
    (WORKSPACE, WORKSPACES),
    (SERVICE_CONTROL_POLICY, SERVICE_CONTROL_POLICIES),
    (SIMULATE_POLICY, SIMULATE_POLICIES),
    (TAG_POLICY, TAG_POLICIES),
    (ORGANIZATIONAL_UNIT, ORGANIZATIONAL_UNITS),
    (C7N_AWS_LAMBDA, C7N_AWS_LAMBDAS),
]

SECTION_NAME_SINGULAR_AND_PLURAL_LIST_THAT_SUPPORTS_PARAMETERS = [
    (LAUNCH, LAUNCHES),
    (STACK, STACKS),
    (LAMBDA_INVOCATION, LAMBDA_INVOCATIONS),
    (CODE_BUILD_RUN, CODE_BUILD_RUNS),
    (APP, APPS),
    (WORKSPACE, WORKSPACES),
]

SECTION_NAMES_THAT_SUPPORTS_PARAMETERS = [
    LAUNCHES,
    STACKS,
    LAMBDA_INVOCATIONS,
    CODE_BUILD_RUNS,
    APPS,
    WORKSPACES,
]

SECTION_SINGULAR_TO_PLURAL = {
    LAUNCH: LAUNCHES,
    STACK: STACKS,
    SPOKE_LOCAL_PORTFOLIO: SPOKE_LOCAL_PORTFOLIOS,
    IMPORTED_PORTFOLIO: IMPORTED_PORTFOLIOS,
    LAMBDA_INVOCATION: LAMBDA_INVOCATIONS,
    CODE_BUILD_RUN: CODE_BUILD_RUNS,
    ASSERTION: ASSERTIONS,
    APP: APPS,
    WORKSPACE: WORKSPACES,
    SIMULATE_POLICY: SIMULATE_POLICIES,
    SERVICE_CONTROL_POLICY: SERVICE_CONTROL_POLICIES,
    C7N_AWS_LAMBDA: C7N_AWS_LAMBDAS,
}

CODEBUILD_DEFAULT_IMAGE = "aws/codebuild/standard:7.0"

DEFAULT_TERRAFORM_VERSION_PARAMETER_NAME = (
    "/servicecatalog-puppet/terraform/default-version"
)
DEFAULT_TERRAFORM_VERSION_VALUE = "1.0.4"

EXECUTE_TERRAFORM_PROJECT_NAME = "servicecatalog-puppet-execute-terraform"
EXECUTE_DRY_RUN_TERRAFORM_PROJECT_NAME = (
    "servicecatalog-puppet-execute-dry-run-terraform"
)
TERMINATE_TERRAFORM_PROJECT_NAME = "servicecatalog-puppet-terminate-terraform"
TERMINATE_DRY_RUN_TERRAFORM_PROJECT_NAME = (
    "servicecatalog-puppet-terminate-dry-run-terraform"
)
TERRAFORM_SPOKE_PREP_STACK_NAME = f"{BOOTSTRAP_STACK_NAME}-terraform-execution"
STACK_SPOKE_PREP_STACK_NAME = f"{BOOTSTRAP_STACK_NAME}-stack-execution"

USE_SERVICE_ROLE_DEFAULT = False

MANIFEST_STATUS_FIELD_NAME = "status"
MANIFEST_STATUS_FIELD_VALUE_IGNORED = "ignored"

CONFIG_SHOULD_DELETE_ROLLBACK_COMPLETE_STACKS = "should_delete_rollback_complete_stacks"
CONFIG_SHOULD_DELETE_ROLLBACK_COMPLETE_STACKS_DEFAULT = False

PUPPET_LOGGER_NAME = "puppet-logger"
PUPPET_SCHEDULER_LOGGER_NAME = "puppet-logger-scheduler"

SPOKE_EXECUTION_MODE_DEPLOY_ENV_PARAMETER_NAME = (
    "/servicecatalog-puppet/spoke/deploy-environment-compute-type"
)
SPOKE_EXECUTION_MODE_DEPLOY_ENV_DEFAULT = "BUILD_GENERAL1_SMALL"

STATIC_HTML_PAGE = "static-html-page.html"

FULL_RUN_CODEBUILD_PROJECT_NAME = "servicecatalog-puppet-deploy"
SINGLE_ACCOUNT_RUN_CODEBUILD_PROJECT_NAME = "servicecatalog-puppet-single-account-run"

INITIALISER_STACK_NAME_SSM_PARAMETER = "service-catalog-puppet-initialiser-stack-name"

SECTIONS_THAT_SUPPORT_CONDITIONS = [
    LAUNCHES,
    SPOKE_LOCAL_PORTFOLIOS,
    IMPORTED_PORTFOLIOS,
    STACKS,
    APPS,
    WORKSPACES,
    LAMBDA_INVOCATIONS,
    CODE_BUILD_RUNS,
    SERVICE_CONTROL_POLICIES,
    TAG_POLICIES,
    SIMULATE_POLICIES,
    ASSERTIONS,
    C7N_AWS_LAMBDAS,
]

SERVICE_CATALOG_PUPPET_MANIFEST_SSM_PREFIX = "/servicecatalog-puppet/manifest"

DEPLOY_TO_NAMES = {
    LAUNCHES: "deploy_to",
    STACKS: "deploy_to",
    SPOKE_LOCAL_PORTFOLIOS: "share_with",
    IMPORTED_PORTFOLIOS: "share_with",
    LAMBDA_INVOCATIONS: "invoke_for",
    CODE_BUILD_RUNS: "run_for",
    ASSERTIONS: "assert_for",
    APPS: "deploy_to",
    WORKSPACES: "deploy_to",
    SERVICE_CONTROL_POLICIES: "apply_to",
    SIMULATE_POLICIES: "simulate_for",
    TAG_POLICIES: "apply_to",
    C7N_AWS_LAMBDAS: "apply_to",
}

CLOUDFORMATION_HAPPY_STATUS = [
    "CREATE_COMPLETE",
    "UPDATE_ROLLBACK_COMPLETE",
    "UPDATE_COMPLETE",
    "IMPORT_COMPLETE",
    "IMPORT_ROLLBACK_COMPLETE",
]

CLOUDFORMATION_UNHAPPY_STATUS = [
    "CREATE_FAILED",
    "ROLLBACK_FAILED",
    "DELETE_FAILED",
    "UPDATE_FAILED",
    "UPDATE_ROLLBACK_FAILED",
    "IMPORT_ROLLBACK_FAILED",
]

CLOUDFORMATION_IN_PROGRESS_STATUS = [
    "CREATE_IN_PROGRESS",
    "ROLLBACK_IN_PROGRESS",
    "DELETE_IN_PROGRESS",
    "UPDATE_COMPLETE_CLEANUP_IN_PROGRESS",
    "UPDATE_IN_PROGRESS",
    "UPDATE_ROLLBACK_IN_PROGRESS",
    "UPDATE_ROLLBACK_COMPLETE_CLEANUP_IN_PROGRESS",
    "REVIEW_IN_PROGRESS",
    "IMPORT_IN_PROGRESS",
    "IMPORT_ROLLBACK_IN_PROGRESS",
]

SSM_OUTPUTS = "ssm_outputs"
SSM_PARAMETERS = "ssm_parameters"
S3_GET_OBJECT = "s3_get_object"
BOTO3_PARAMETERS = "boto3_parameters"
SSM_PARAMETERS_WITH_A_PATH = "ssm_parameters_with_a_path"
PORTFOLIO_ASSOCIATIONS = "portfolio-associations"
PORTFOLIO_CONSTRAINTS_LAUNCH = "portfolio-constraints-launch"
PORTFOLIO_CONSTRAINTS_RESOURCE_UPDATE = "portfolio-constraints-resource_update"
PORTFOLIO_COPY = "portfolio-copy"
PORTFOLIO_IMPORT = "portfolio-import"
PORTFOLIO_SHARE_AND_ACCEPT_ACCOUNT = "portfolio-share-and-accept-account"
PORTFOLIO_SHARE_AND_ACCEPT_AWS_ORGANIZATIONS = (
    "portfolio-share-and-accept-aws_organizations"
)
PORTFOLIO_GET_ALL_PRODUCTS_AND_THEIR_VERSIONS = (
    "portfolio-get-all-products-and-their-versions"
)
PORTFOLIO_DISASSOCIATE_ALL_PRODUCTS_AND_THEIR_VERSIONS = (
    "portfolio-disassociate-all-products-and-their-versions"
)
PORTFOLIO_LOCAL = "portfolio-local"
PORTFOLIO_IMPORTED = "portfolio-imported"
DESCRIBE_PROVISIONING_PARAMETERS = "describe-provisioning-parameters"
PORTFOLIO_PUPPET_ROLE_ASSOCIATION = "portfolio-puppet-role-association"
WORKSPACE_ACCOUNT_PREPARATION = "workspace-account-preparation"
RUN_DEPLOY_IN_SPOKE = "run-deploy-in-spoke"
GENERATE_MANIFEST = "generate-manifest"
GET_TEMPLATE_FROM_S3 = "get-template-from-s3"
GET_OR_CREATE_SERVICE_CONTROL_POLICIES_POLICY = (
    "get-or-create-service-control-policies-policy"
)
GET_OR_CREATE_TAG_POLICIES_POLICY = "get-or-create-tag-policies-policy"

PREPARE_ACCOUNT_FOR_STACKS = "prepare-account-for-stacks"
CREATE_POLICIES = "create-policies"

PARAMETERS_TO_TRY_AS_OPERATIONAL_DATA = [
    "puppet_account_id",
    "account_id",
    "region",
    "task_reference",
]

RUN_DEPLOY_IN_SPOKE_BUILDSPEC = """
      version: 0.2
      phases:
        install:
          runtime-versions:
            python: 3.11
          commands:
            - {}
        build:
          commands:
            - curl $TASK_REFERENCE_URL > manifest-task-reference.json
            - curl $MANIFEST_URL > manifest-expanded.yaml
            - servicecatalog-puppet --info deploy-in-spoke-from-task-reference --execution-mode spoke --puppet-account-id $PUPPET_ACCOUNT_ID --single-account $(aws sts get-caller-identity --query Account --output text) --home-region $HOME_REGION --regions $REGIONS --should-collect-cloudformation-events $SHOULD_COLLECT_CLOUDFORMATION_EVENTS --should-forward-events-to-eventbridge $SHOULD_FORWARD_EVENTS_TO_EVENTBRIDGE --should-forward-failures-to-opscenter $SHOULD_FORWARD_FAILURES_TO_OPSCENTER ${{PWD}}
      artifacts:
        files:
          - results/*/*
          - output/*/*
        name: DeployInSpokeProject
"""

SCHEDULER_THREADS_OR_PROCESSES_DEFAULT = "threads"

REPORTING_ROLE_NAME = "PuppetRoleForReporting"
CACHE_DOWNLOADING_ROLE_NAME = "PuppetRoleForDownloadingFromCache"

SHARE_PRINCIPALS_DEFAULT = False
AWS_STS_REGIONAL_ENDPOINTS_LEGACY = "legacy"
AWS_STS_REGIONAL_ENDPOINTS_REGIONAL = "regional"
AWS_STS_REGIONAL_ENDPOINTS_DEFAULT = AWS_STS_REGIONAL_ENDPOINTS_LEGACY
DESCRIBE_PORTFOLIO_SHARES = "describe-portfolio-shares"

SCHEDULER_ALGORITHM_DEFAULT = "topological_generations"
SPOKE_SCHEDULER_THREADS_OR_PROCESSES_DEFAULT = SCHEDULER_THREADS_OR_PROCESSES_DEFAULT
SPOKE_SCHEDULER_ALGORITHM_DEFAULT = SCHEDULER_ALGORITHM_DEFAULT

CACHE_LEVEL_RUN = "CACHE_LEVEL_RUN"
CACHE_LEVEL_TASK = "CACHE_LEVEL_TASK"
CACHE_LEVEL_PERMANENT = "CACHE_LEVEL_PERMANENT"
CACHE_LEVEL_NO_CACHE = "CACHE_LEVEL_NO_CACHE"

CACHE_LEVEL_DEFAULT = CACHE_LEVEL_RUN

WAS_A_SPOKE_LOCAL_PORTFOLIO_DEFAULT = False

C7N_ORGS_INPUT_FILE_TASK = "c7n_orgs_input_file_task"
C7N_FORWARD_EVENTS_FOR_ACCOUNT_TASK = "c7n_forward_events_for_account_task"
C7N_FORWARD_EVENTS_FOR_REGION_TASK = "c7n_forward_events_for_region_task"
C7N_DEPLOY_POLICIES_TASK = "c7n_deploy_policies_task"
C7N_PREPARE_HUB_ACCOUNT_TASK = "c7n_prepare_hub_account_task"
C7N_CREATE_CUSTODIAN_ROLE_TASK = "c7n_create_custodian_role_task"
C7N_CUSTODIAN_ROLE_NAME_DEFAULT = "Custodian"
C7N_CUSTODIAN_ROLE_PATH_DEFAULT = "/servicecatalog-puppet/c7n/"
C7N_CUSTODIAN_MANAGED_POLICY_ARNS_DEFAULT = [
    "arn:aws:iam::${AWS::Partition}:policy/AdministratorAccess"
]
C7N_VERSION_DEFAULT = "0.9.28"
TERMINATE_CLOUDFORMATION_STACK_TASK = "terminate-CloudFormation-stack"



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0
import unittest
from unittest import mock

from servicecatalog_puppet import constants


class TestUsersOfGetConfig(unittest.TestCase):
    puppet_account_id = "abcd"
    default_region = "eu-west-0"

    def setUp(self) -> None:
        from servicecatalog_puppet import remote_config

        self.module = remote_config

    def test_get_should_simple_values(self):
        # setup
        expected_result = "banananana"
        checks_to_perform = [
            (
                self.module.get_should_use_sns,
                constants.CONFIG_SHOULD_COLLECT_CLOUDFORMATION_EVENTS,
            ),
            (
                self.module.get_should_delete_rollback_complete_stacks,
                constants.CONFIG_SHOULD_DELETE_ROLLBACK_COMPLETE_STACKS,
            ),
            (self.module.get_should_use_product_plans, "should_use_product_plans"),
            (
                self.module.get_should_use_eventbridge,
                "should_forward_events_to_eventbridge",
            ),
            (
                self.module.get_should_forward_failures_to_opscenter,
                "should_forward_failures_to_opscenter",
            ),
            (self.module.get_drift_token, "drift_token"),
            (self.module.get_regions, constants.CONFIG_REGIONS),
            (self.module.is_caching_enabled, "is_caching_enabled"),
            (
                self.module.get_global_sharing_mode_default,
                "global_sharing_mode_default",
            ),
            (
                self.module.get_global_share_tag_options_default,
                "global_share_tag_options_default",
            ),
            (
                self.module.get_global_share_principals_default,
                "global_share_principals_default",
            ),
            (
                self.module.get_spoke_deploy_environment_compute_type,
                "spoke_deploy_environment_compute_type",
            ),
            (
                self.module.get_scheduler_threads_or_processes,
                "scheduler_threads_or_processes",
            ),
            (self.module.get_scheduler_algorithm, "scheduler_algorithm"),
        ]
        with mock.patch(
            "servicecatalog_puppet.remote_config.get_config"
        ) as mocked_get_config:
            for method_to_check, config_key_value in checks_to_perform:
                with self.subTest(
                    method_to_check=method_to_check, config_key_value=config_key_value
                ):
                    mocked_get_config.return_value = {config_key_value: expected_result}
                    # exercise
                    actual_result = method_to_check(
                        self.puppet_account_id, self.default_region
                    )
                    # verify
                    self.assertEquals(expected_result, actual_result)



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0
import unittest
from copy import deepcopy

from servicecatalog_puppet import constants


class TestManifestForExplode(unittest.TestCase):
    account_a = {
        "account_id": "012345678910",
        "default_region": "eu-west-1",
        "name": "accounta",
        "expanded_from": "ou-aaaa-aaaaaaaa",
        "organization": "o-aaaaaaaa",
        "regions_enabled": [
            "eu-west-2",
        ],
        "tags": ["group:A"],
    }
    account_b = {
        "account_id": "009876543210",
        "default_region": "us-west-1",
        "expanded_from": "ou-bbbb-bbbbbbbb",
        "organization": "o-bbbbbbbb",
        "name": "accountb",
        "regions_enabled": [
            "us-west-2",
        ],
        "tags": ["group:B"],
    }
    account_c = {
        "account_id": "432100098765",
        "default_region": "ap-west-1",
        "name": "accountc",
        "regions_enabled": [
            "ap-west-2",
        ],
        "tags": ["group:C"],
    }
    accounts = [
        account_a,
        account_b,
        account_c,
    ]

    lambda_invocation_with_no_dependencies = dict(
        product="lambda_invocation_with_no_dependencies",
    )
    lambda_invocation_with_one_lambda_dependency = dict(
        function_name="lambda_invocation_with_one_lambda_dependency",
        depends_on=[
            dict(
                name="lambda_invocation_with_no_dependencies",
                type=constants.LAMBDA_INVOCATION,
            )
        ],
    )
    lambda_invocation_with_one_launch_dependency = dict(
        function_name="lambda_invocation_with_one_launch_dependency",
        depends_on=[
            dict(
                name="launch_with_no_dependencies",
                type=constants.LAUNCH,
            )
        ],
    )
    launch_with_no_dependencies = dict(
        product="launch_with_no_dependencies",
    )
    spoke_local_portfolio_with_no_dependencies = dict(
        product="spoke_local_portfolio_with_no_dependencies",
    )
    spoke_local_portfolio_with_one_launch_dependency = dict(
        product="spoke_local_portfolio_with_one_launch_dependency",
        depends_on=["launch_with_no_dependencies"],
    )
    launch_depending_on_depending_launch = dict(
        product="launch_depending_on_depending_launch",
        depends_on=[dict(type=constants.LAUNCH, name="launch_depending_on_launch")],
    )
    launch_depending_on_launch = dict(
        product="launch_depending_on_launch",
        depends_on=[dict(type=constants.LAUNCH, name="launch_with_no_dependencies")],
    )
    another_launch_with_no_dependencies = dict(
        product="another_launch_with_no_dependencies",
    )
    launch_with_one_lambda_dependency = dict(
        function_name="launch_with_one_lambda_dependency",
        depends_on=["launch_with_no_dependencies"],
    )
    launch_depending_on_lambda = dict(
        function_name="launch_depending_on_lambda",
        depends_on=[
            dict(
                name="lambda_invocation_with_no_dependencies",
                type=constants.LAMBDA_INVOCATION,
            )
        ],
    )

    def setUp(self):
        self.maxDiff = None
        from servicecatalog_puppet import manifest_utils

        self.sut = manifest_utils.explode

    def tearDown(self):
        self.sut = None

    def test_explode_simple_for_lambda_invoke(self):
        # setup
        expanded_manifest = {
            constants.LAUNCHES: dict(),
            constants.STACKS: dict(),
            constants.SPOKE_LOCAL_PORTFOLIOS: dict(),
            constants.LAMBDA_INVOCATIONS: dict(
                lambda_invocation_with_no_dependencies=self.lambda_invocation_with_no_dependencies
            ),
            constants.ASSERTIONS: dict(),
            constants.SERVICE_CONTROL_POLICIES: dict(),
            constants.TAG_POLICIES: dict(),
            constants.APPS: dict(),
            constants.WORKSPACES: dict(),
        }

        expected_results = [expanded_manifest]

        # exercise
        actual_results = self.sut(expanded_manifest)

        # verify
        self.assertListEqual(actual_results, expected_results)

    def test_explode_simple_for_launch(self):
        # setup

        expanded_manifest = {
            constants.LAUNCHES: dict(
                launch_with_no_dependencies=self.launch_with_no_dependencies
            ),
            constants.STACKS: dict(),
            constants.SPOKE_LOCAL_PORTFOLIOS: dict(),
            constants.LAMBDA_INVOCATIONS: dict(),
            constants.ASSERTIONS: dict(),
            constants.SERVICE_CONTROL_POLICIES: dict(),
            constants.TAG_POLICIES: dict(),
            constants.APPS: dict(),
            constants.WORKSPACES: dict(),
        }

        expected_results = [expanded_manifest]

        # exercise
        actual_results = self.sut(expanded_manifest)

        # verify
        self.assertListEqual(actual_results, expected_results)

    def test_explode_simple_for_spoke_local_portfolio(self):
        # setup
        expanded_manifest = {
            constants.LAUNCHES: dict(),
            constants.STACKS: dict(),
            constants.SPOKE_LOCAL_PORTFOLIOS: dict(
                spoke_local_portfolio_with_no_dependencies=self.spoke_local_portfolio_with_no_dependencies
            ),
            constants.LAMBDA_INVOCATIONS: dict(),
            constants.ASSERTIONS: dict(),
            constants.SERVICE_CONTROL_POLICIES: dict(),
            constants.TAG_POLICIES: dict(),
            constants.APPS: dict(),
            constants.WORKSPACES: dict(),
        }

        expected_results = [expanded_manifest]

        # exercise
        actual_results = self.sut(expanded_manifest)

        # verify
        self.assertListEqual(actual_results, expected_results)

    def test_explode_simple_for_spoke_local_portfolio_depends_on_launch(self):
        # setup
        expanded_manifest = {
            constants.LAUNCHES: dict(
                launch_with_no_dependencies=self.launch_with_no_dependencies,
            ),
            constants.STACKS: dict(),
            constants.SPOKE_LOCAL_PORTFOLIOS: dict(
                spoke_local_portfolio_with_no_dependencies=self.spoke_local_portfolio_with_one_launch_dependency,
            ),
            constants.LAMBDA_INVOCATIONS: dict(),
            constants.ASSERTIONS: dict(),
            constants.SERVICE_CONTROL_POLICIES: dict(),
            constants.TAG_POLICIES: dict(),
            constants.APPS: dict(),
            constants.WORKSPACES: dict(),
        }

        expected_results = [expanded_manifest]

        # exercise
        actual_results = self.sut(expanded_manifest)

        # verify
        self.assertListEqual(actual_results, expected_results)

    def test_explode_simple_for_two_launches(self):
        # setup
        expanded_manifest = {
            constants.LAUNCHES: dict(
                launch_with_no_dependencies=self.launch_with_no_dependencies,
                another_launch_with_no_dependencies=self.another_launch_with_no_dependencies,
            ),
            constants.STACKS: dict(),
            constants.SPOKE_LOCAL_PORTFOLIOS: dict(),
            constants.LAMBDA_INVOCATIONS: dict(),
            constants.ASSERTIONS: dict(),
            constants.SERVICE_CONTROL_POLICIES: dict(),
            constants.TAG_POLICIES: dict(),
            constants.APPS: dict(),
            constants.WORKSPACES: dict(),
        }

        expected_results_1 = deepcopy(expanded_manifest)
        expected_results_1[constants.LAUNCHES] = dict(
            launch_with_no_dependencies=self.launch_with_no_dependencies,
        )
        expected_results_2 = deepcopy(expanded_manifest)
        expected_results_2[constants.LAUNCHES] = dict(
            another_launch_with_no_dependencies=self.another_launch_with_no_dependencies,
        )
        expected_results = [expected_results_1, expected_results_2]

        # exercise
        actual_results = self.sut(expanded_manifest)

        # verify
        self.assertListEqual(actual_results, expected_results)

    def test_explode_single_lambda_dependency_on_lambda(self):
        # setup
        expanded_manifest = {
            constants.LAUNCHES: dict(),
            constants.STACKS: dict(),
            constants.SPOKE_LOCAL_PORTFOLIOS: dict(),
            constants.LAMBDA_INVOCATIONS: dict(
                lambda_invocation_with_no_dependencies=self.lambda_invocation_with_no_dependencies,
                lambda_invocation_with_one_lambda_dependency=self.lambda_invocation_with_one_lambda_dependency,
            ),
            constants.ASSERTIONS: dict(),
            constants.SERVICE_CONTROL_POLICIES: dict(),
            constants.TAG_POLICIES: dict(),
            constants.APPS: dict(),
            constants.WORKSPACES: dict(),
        }

        expected_results = [expanded_manifest]

        # exercise
        actual_results = self.sut(expanded_manifest)

        # verify
        self.assertListEqual(actual_results, expected_results)

    def test_explode_single_lambda_dependency_on_launch(self):
        # setup
        expanded_manifest = {
            constants.LAUNCHES: dict(
                launch_with_no_dependencies=self.launch_with_no_dependencies
            ),
            constants.STACKS: dict(),
            constants.SPOKE_LOCAL_PORTFOLIOS: dict(),
            constants.LAMBDA_INVOCATIONS: dict(
                lambda_invocation_with_one_launch_dependency=self.lambda_invocation_with_one_launch_dependency,
            ),
            constants.ASSERTIONS: dict(),
            constants.SERVICE_CONTROL_POLICIES: dict(),
            constants.TAG_POLICIES: dict(),
            constants.APPS: dict(),
            constants.WORKSPACES: dict(),
        }

        expected_results = [expanded_manifest]

        # exercise
        actual_results = self.sut(expanded_manifest)

        # verify
        self.assertListEqual(actual_results, expected_results)

    def test_explode_single_launch_depends_on_lambda(self):
        # setup
        expanded_manifest = {
            constants.LAUNCHES: dict(
                launch_depending_on_lambda=self.launch_depending_on_lambda
            ),
            constants.STACKS: dict(),
            constants.SPOKE_LOCAL_PORTFOLIOS: dict(),
            constants.LAMBDA_INVOCATIONS: dict(
                lambda_invocation_with_no_dependencies=self.lambda_invocation_with_no_dependencies,
            ),
            constants.ASSERTIONS: dict(),
            constants.SERVICE_CONTROL_POLICIES: dict(),
            constants.TAG_POLICIES: dict(),
            constants.APPS: dict(),
            constants.WORKSPACES: dict(),
        }

        expected_results = [expanded_manifest]

        # exercise
        actual_results = self.sut(expanded_manifest)

        # verify
        self.assertListEqual(actual_results, expected_results)

    def test_explode_single_lambda_dependency_on_lambda_reversed(self):
        # setup
        expanded_manifest = {
            constants.LAUNCHES: dict(),
            constants.STACKS: dict(),
            constants.SPOKE_LOCAL_PORTFOLIOS: dict(),
            constants.LAMBDA_INVOCATIONS: dict(
                lambda_invocation_with_one_lambda_dependency=self.lambda_invocation_with_one_lambda_dependency,
                lambda_invocation_with_no_dependencies=self.lambda_invocation_with_no_dependencies,
            ),
            constants.ASSERTIONS: dict(),
            constants.SERVICE_CONTROL_POLICIES: dict(),
            constants.TAG_POLICIES: dict(),
            constants.APPS: dict(),
            constants.WORKSPACES: dict(),
        }

        expected_results = [expanded_manifest]

        # exercise
        actual_results = self.sut(expanded_manifest)

        # verify
        self.assertListEqual(actual_results, expected_results)

    def test_explode_single_launch_dependency_on_launch(self):
        # setup
        expanded_manifest = {
            constants.LAUNCHES: dict(
                launch_with_no_dependencies=self.launch_with_no_dependencies,
                launch_with_one_lambda_dependency=self.launch_with_one_lambda_dependency,
            ),
            constants.STACKS: dict(),
            constants.SPOKE_LOCAL_PORTFOLIOS: dict(),
            constants.LAMBDA_INVOCATIONS: dict(),
            constants.ASSERTIONS: dict(),
            constants.SERVICE_CONTROL_POLICIES: dict(),
            constants.TAG_POLICIES: dict(),
            constants.APPS: dict(),
            constants.WORKSPACES: dict(),
        }

        expected_results = [expanded_manifest]

        # exercise
        actual_results = self.sut(expanded_manifest)

        # verify
        self.assertListEqual(actual_results, expected_results)

    def test_explode_single_launch_dependency_on_launch_reversed(self):
        # setup
        expanded_manifest = {
            constants.LAUNCHES: dict(
                launch_with_one_lambda_dependency=self.launch_with_one_lambda_dependency,
                launch_with_no_dependencies=self.launch_with_no_dependencies,
            ),
            constants.STACKS: dict(),
            constants.SPOKE_LOCAL_PORTFOLIOS: dict(),
            constants.LAMBDA_INVOCATIONS: dict(),
            constants.ASSERTIONS: dict(),
            constants.SERVICE_CONTROL_POLICIES: dict(),
            constants.TAG_POLICIES: dict(),
            constants.APPS: dict(),
            constants.WORKSPACES: dict(),
        }

        expected_results = [expanded_manifest]

        # exercise
        actual_results = self.sut(expanded_manifest)

        # verify
        self.assertListEqual(actual_results, expected_results)

    def test_explode_3_launches_with_a_gap(self):
        # setup
        expanded_manifest = {
            constants.LAUNCHES: dict(
                launch_depending_on_depending_launch=self.launch_depending_on_depending_launch,
                launch_with_no_dependencies=self.launch_with_no_dependencies,
                launch_depending_on_launch=self.launch_depending_on_launch,
            ),
            constants.STACKS: dict(),
            constants.SPOKE_LOCAL_PORTFOLIOS: dict(),
            constants.LAMBDA_INVOCATIONS: dict(),
            constants.ASSERTIONS: dict(),
            constants.SERVICE_CONTROL_POLICIES: dict(),
            constants.TAG_POLICIES: dict(),
            constants.APPS: dict(),
            constants.WORKSPACES: dict(),
        }

        expected_results = [expanded_manifest]

        # exercise
        actual_results = self.sut(expanded_manifest)

        # verify
        self.assertListEqual(actual_results, expected_results)



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0
from servicecatalog_puppet import constants, task_reference_constants


def handle_workspaces(
    all_tasks,
    all_tasks_task_reference,
    item_name,
    puppet_account_id,
    section_name,
    task_reference,
    task_to_add,
):
    workspace_account_preparation_ref = (
        f"{constants.WORKSPACE_ACCOUNT_PREPARATION}-{task_to_add.get('account_id')}"
    )
    if all_tasks.get(workspace_account_preparation_ref) is None:
        all_tasks[workspace_account_preparation_ref] = {
            "puppet_account_id": puppet_account_id,
            "task_reference": workspace_account_preparation_ref,
            "dependencies_by_reference": [constants.CREATE_POLICIES],
            "account_id": task_to_add.get("account_id"),
            "section_name": constants.WORKSPACE_ACCOUNT_PREPARATION,
            task_reference_constants.MANIFEST_SECTION_NAMES: dict(),
            task_reference_constants.MANIFEST_ITEM_NAMES: dict(),
            task_reference_constants.MANIFEST_ACCOUNT_IDS: dict(),
        }
    all_tasks[workspace_account_preparation_ref][
        task_reference_constants.MANIFEST_SECTION_NAMES
    ].update(task_to_add.get(task_reference_constants.MANIFEST_SECTION_NAMES))
    all_tasks[workspace_account_preparation_ref][
        task_reference_constants.MANIFEST_ITEM_NAMES
    ].update(task_to_add.get(task_reference_constants.MANIFEST_ITEM_NAMES))
    all_tasks[workspace_account_preparation_ref][
        task_reference_constants.MANIFEST_ACCOUNT_IDS
    ].update(task_to_add.get(task_reference_constants.MANIFEST_ACCOUNT_IDS))
    if workspace_account_preparation_ref not in all_tasks[all_tasks_task_reference].get(
        "dependencies_by_reference"
    ):
        all_tasks[all_tasks_task_reference]["dependencies_by_reference"].append(
            workspace_account_preparation_ref
        )



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import os

from servicecatalog_puppet import (
    constants,
    environmental_variables,
    serialisation_utils,
)


class EnvVarMixin:
    @property
    def drift_token(self):
        return os.environ.get(environmental_variables.DRIFT_TOKEN, "NOW")

    @property
    def run_token(self):
        return os.environ.get(environmental_variables.RUN_TOKEN, "NOW")

    @property
    def should_use_product_plans(self):
        if self.execution_mode == constants.EXECUTION_MODE_HUB:
            return (
                os.environ.get(environmental_variables.SHOULD_USE_PRODUCT_PLANS, "True")
                == "True"
            )
        else:
            return False

    @property
    def single_account(self):
        return os.environ.get("SCT_SINGLE_ACCOUNT", "None")

    @property
    def should_delete_rollback_complete_stacks(self):
        return (
            str(
                os.environ.get(
                    environmental_variables.SHOULD_DELETE_ROLLBACK_COMPLETE_STACKS,
                    constants.CONFIG_SHOULD_DELETE_ROLLBACK_COMPLETE_STACKS_DEFAULT,
                )
            ).upper()
            == "TRUE"
        )

    @property
    def initialiser_stack_tags(self):
        initialiser_stack_tags_value = os.environ.get(
            environmental_variables.INITIALISER_STACK_TAGS, None
        )
        if initialiser_stack_tags_value is None:
            raise Exception(
                f"You must export {environmental_variables.INITIALISER_STACK_TAGS}"
            )
        return serialisation_utils.json_loads(initialiser_stack_tags_value)

    @property
    def spoke_execution_mode_deploy_env(self):
        return os.environ.get(
            environmental_variables.SPOKE_EXECUTION_MODE_DEPLOY_ENV,
            constants.SPOKE_EXECUTION_MODE_DEPLOY_ENV_DEFAULT,
        )

    @property
    def should_use_sns(self):
        return os.environ.get(environmental_variables.SHOULD_USE_SNS, "False") == "True"

    @property
    def is_dry_run(self):
        return os.environ.get("SCT_IS_DRY_RUN", "False") == "True"

    @property
    def executor_account_id(self):
        return os.environ.get(environmental_variables.EXECUTOR_ACCOUNT_ID)

    @property
    def execution_mode(self):
        return os.environ.get(
            environmental_variables.EXECUTION_MODE, constants.EXECUTION_MODE_HUB
        )



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import luigi
import troposphere as t
from troposphere import events

from servicecatalog_puppet import config, constants
from servicecatalog_puppet.workflow.dependencies import tasks


c7nTrailBucket = "c7nTrailBucket"


class ForwardEventsForRegionTask(tasks.TaskWithReferenceAndCommonParameters):
    c7n_account_id = luigi.Parameter()
    custodian_region = luigi.Parameter()
    cachable_level = constants.CACHE_LEVEL_RUN

    def params_for_results_display(self):
        return {
            "task_reference": self.task_reference,
            "region": self.region,
            "account_id": self.account_id,
        }

    def run(self):
        tpl = t.Template()
        tpl.description = (
            "event forwarder template for c7n created by service catalog puppet"
        )
        tpl.add_resource(
            events.Rule(
                "ForwardAll",
                Description="Forward all events for c7n to do its job",
                EventPattern={
                    "account": [self.account_id],
                },
                State="ENABLED",
                Targets=[
                    events.Target(
                        Arn=t.Sub(
                            "arn:${AWS::Partition}:events:"
                            + self.custodian_region
                            + ":"
                            + self.c7n_account_id
                            + ":event-bus/default"
                        ),
                        Id="CloudCustodianHubEventBusArn",
                        RoleArn=t.Sub(
                            "arn:${AWS::Partition}:iam::${AWS::AccountId}:role/servicecatalog-puppet/c7n/c7nEventForwarder"
                        ),
                    )
                ],
            )
        )
        template = tpl.to_yaml()
        with self.spoke_regional_client("cloudformation") as cloudformation:
            cloudformation.create_or_update(
                Capabilities=["CAPABILITY_NAMED_IAM"],
                ShouldUseChangeSets=False,
                StackName="servicecatalog-puppet-c7n-eventforwarding-region",
                TemplateBody=template,
                NotificationARNs=(
                    [
                        f"arn:{config.get_partition()}:sns:{self.region}:{self.puppet_account_id}:servicecatalog-puppet-cloudformation-regional-events"
                    ]
                    if self.should_use_sns
                    else []
                ),
                ShouldDeleteRollbackComplete=self.should_delete_rollback_complete_stacks,
                Tags=self.initialiser_stack_tags,
            )
        self.write_empty_output()



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

from servicecatalog_puppet.waluigi.processes import runner as processes_runner
from servicecatalog_puppet.waluigi.threads import runner as threads_runner


def get_runner(threads_or_processes: str):
    if threads_or_processes == "threads":
        return threads_runner
    elif threads_or_processes == "processes":
        return processes_runner

    raise ValueError(f"threads_or_processes invalid: {threads_or_processes}")



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0
import json
import logging
import os

import click
import networkx as nx
import requests
import yamale
import yaml
from betterboto import client as betterboto_client

from servicecatalog_puppet import (
    asset_helpers,
    config,
    constants,
    manifest_utils,
    print_utils,
    serialisation_utils,
)


logger = logging.getLogger(__name__)


def assemble_manifest_from_ssm(target_directory):
    with betterboto_client.ClientContextManager("ssm") as ssm:
        paginator = ssm.get_paginator("get_parameters_by_path")
        manifest = {
            "schema": "puppet-2019-04-01",
            constants.LAUNCHES: dict(),
            constants.STACKS: dict(),
            constants.SPOKE_LOCAL_PORTFOLIOS: dict(),
            constants.ASSERTIONS: {},
            constants.CODE_BUILD_RUNS: {},
            constants.LAMBDA_INVOCATIONS: {},
            constants.APPS: {},
            constants.WORKSPACES: {},
            constants.CFCT: {},
            constants.SERVICE_CONTROL_POLICIES: {},
            constants.SIMULATE_POLICIES: {},
            constants.TAG_POLICIES: {},
        }
        for page in paginator.paginate(
            Path=constants.SERVICE_CATALOG_PUPPET_MANIFEST_SSM_PREFIX,
            Recursive=True,
        ):
            for parameter in page.get("Parameters", []):
                parts = parameter.get("Name").split("/")
                action_type = parts[3]
                action_name = parts[4]
                manifest[action_type][action_name] = serialisation_utils.load(
                    parameter.get("Value")
                )
        if not os.path.exists(target_directory):
            os.makedirs(target_directory)
        open(f"{target_directory}{os.path.sep}ssm_manifest.yaml", "w").write(
            serialisation_utils.dump(manifest)
        )


def expand(f, puppet_account_id, regions, single_account, subset=None):
    click.echo("Expanding")
    target_directory = os.path.sep.join([os.path.dirname(f.name), "manifests"])
    assemble_manifest_from_ssm(target_directory)
    manifest = manifest_utils.load(f, puppet_account_id)
    org_iam_role_arn = config.get_org_iam_role_arn(puppet_account_id)
    if org_iam_role_arn is None:
        click.echo("No org role set - not expanding")
        new_manifest = manifest
    else:
        click.echo("Expanding using role: {}".format(org_iam_role_arn))
        with betterboto_client.CrossAccountClientContextManager(
            "organizations", org_iam_role_arn, "org-iam-role"
        ) as client:
            conversions, new_manifest = manifest_utils.expand_manifest(manifest, client)
            new_manifest = manifest_utils.rewrite_organizational_units(
                new_manifest, conversions, client
            )
    click.echo("Expanded")

    new_manifest = manifest_utils.rewrite_deploy_as_share_to_for_spoke_local_portfolios(
        new_manifest
    )
    if single_account:
        click.echo(f"Filtering for single account: {single_account}")

        for account in new_manifest.get("accounts", []):
            if str(account.get("account_id")) == str(single_account):
                click.echo(f"Found single account: {single_account}")
                new_manifest["accounts"] = [account]
                break

        items_to_delete = list()
        for section_name in constants.ALL_SECTION_NAMES:
            deploy_to_name = constants.DEPLOY_TO_NAMES[section_name]
            for item_name, item in new_manifest.get(section_name, {}).items():
                accounts = list()
                for deploy_details in item.get(deploy_to_name, {}).get("accounts", []):
                    if str(deploy_details.get("account_id")) == str(single_account):
                        accounts.append(deploy_details)

                if item.get(deploy_to_name).get("accounts"):
                    if len(accounts) > 0:
                        item[deploy_to_name]["accounts"] = accounts
                    else:
                        if item[deploy_to_name].get("tags") or item[deploy_to_name].get(
                            "ous"
                        ):
                            del item[deploy_to_name]["accounts"]
                        else:
                            items_to_delete.append(f"{section_name}:{item_name}")
        for item_to_delete in items_to_delete:
            section_name, item_name = item_to_delete.split(":")
            del new_manifest[section_name][item_name]

        click.echo("Filtered")

    new_manifest = manifest_utils.rewrite_cfct(new_manifest)
    new_manifest = manifest_utils.rewrite_depends_on(new_manifest)
    new_manifest = manifest_utils.rewrite_ssm_parameters(new_manifest)
    new_manifest = manifest_utils.rewrite_stacks(new_manifest, puppet_account_id)
    new_manifest = manifest_utils.rewrite_scps(new_manifest, puppet_account_id)
    new_manifest = manifest_utils.parse_conditions(new_manifest)
    new_manifest = manifest_utils.rewrite_c7n_cloudtrails(new_manifest)

    if subset and subset.get("single_action_section"):
        click.echo(f"Filtering for subset: {subset}")
        new_manifest = manifest_utils.isolate(new_manifest, subset)

    manifest_accounts_all = [
        {"account_id": a.get("account_id"), "email": a.get("email")}
        for a in new_manifest.get("accounts", [])
    ]
    manifest_accounts_excluding = [
        a for a in manifest_accounts_all if a.get("account_id") != puppet_account_id
    ]

    # handle all accounts
    sct_manifest_accounts = json.dumps(manifest_accounts_all)
    sct_manifest_spokes = json.dumps(manifest_accounts_excluding)
    sct_config_regions = json.dumps(regions)

    new_manifest["parameters"]["SCTManifestAccounts"] = dict(
        default=sct_manifest_accounts
    )
    new_manifest["parameters"]["SCTManifestSpokes"] = dict(default=sct_manifest_spokes)
    new_manifest["parameters"]["SCTConfigRegions"] = dict(default=sct_config_regions)
    new_manifest["parameters"]["SCTAccountId"] = dict(default=puppet_account_id)

    if new_manifest.get(constants.LAMBDA_INVOCATIONS) is None:
        new_manifest[constants.LAMBDA_INVOCATIONS] = dict()

    new_name = f.name.replace(".yaml", "-expanded.yaml")
    logger.info("Writing new manifest: {}".format(new_name))
    with open(new_name, "w") as output:
        output.write(serialisation_utils.dump(new_manifest))


def explode(f):
    logger.info("Exploding")
    puppet_account_id = config.get_puppet_account_id()
    original_name = f.name
    expanded_output = f.name.replace(".yaml", "-expanded.yaml")
    expanded_manifest = manifest_utils.load(
        open(expanded_output, "r"), puppet_account_id
    )
    expanded_manifest = manifest_utils.Manifest(expanded_manifest)

    exploded = manifest_utils.explode(expanded_manifest)
    logger.info(f"found {len(exploded)} graphs")
    count = 0
    for mani in exploded:
        with open(original_name.replace(".yaml", f"-exploded-{count}.yaml"), "w") as f:
            f.write(serialisation_utils.dump(mani))
        count += 1


def validate(f):
    logger.info("Validating {}".format(f.name))

    manifest = manifest_utils.load(f, config.get_puppet_account_id())

    try:
        Loader = yaml.CSafeLoader
    except AttributeError:  # System does not have libyaml
        Loader = yaml.SafeLoader
    Loader.add_constructor("!Equals", serialisation_utils.Equals.from_yaml)
    Loader.add_constructor("!Not", serialisation_utils.Not.from_yaml)

    schema = yamale.make_schema(asset_helpers.resolve_from_site_packages("schema.yaml"))
    data = yamale.make_data(content=serialisation_utils.dump(manifest))

    yamale.validate(schema, data, strict=False)

    has_default_default_region = (
        manifest.get("defaults", {}).get("accounts", {}).get("default_region", False)
    )
    has_default_regions_enabled = (
        manifest.get("defaults", {}).get("accounts", {}).get("regions_enabled", False)
    )

    graph = nx.DiGraph()

    tags_defined_by_accounts = {}
    for account in manifest.get("accounts"):
        account_entry_is_an_overwrite_or_append = account.get(
            "append", account.get("overwrite", False)
        )
        if (
            not account_entry_is_an_overwrite_or_append
            and account.get("default_region") is None
            and not has_default_default_region
        ):
            raise Exception(
                f"account entry {account.get('account_id', account.get('ou'))} is missing default_region"
            )
        if (
            not account_entry_is_an_overwrite_or_append
            and account.get("regions_enabled") is None
            and not has_default_regions_enabled
        ):
            raise Exception(
                f"account entry {account.get('account_id', account.get('ou'))} is missing regions_enabled"
            )
        for tag in account.get("tags", []):
            tags_defined_by_accounts[tag] = True
        for tag in account.get("append", {}).get("tags", []):
            tags_defined_by_accounts[tag] = True
        for tag in account.get("overwrite", {}).get("tags", []):
            tags_defined_by_accounts[tag] = True

    for collection_type in constants.ALL_SECTION_NAMES:
        collection_to_check = manifest.get(collection_type, {})
        for collection_name, collection_item in collection_to_check.items():
            #
            # Check the tags in deploy_to that are not defined
            #
            for deploy_to in collection_item.get("deploy_to", {}).get("tags", []):
                tag_to_check = deploy_to.get("tag")
                if tags_defined_by_accounts.get(tag_to_check) is None:
                    print_utils.warn(
                        f"{collection_type}.{collection_name} uses tag {tag_to_check} in deploy_to that does not exist",
                    )

            #
            # Check the depends_on where the dependency is not present
            #
            for depends_on in collection_item.get("depends_on", []):
                if isinstance(depends_on, str):
                    if manifest.get(constants.LAUNCHES).get(depends_on) is None:
                        print_utils.warn(
                            f"{collection_type}.{collection_name} uses {depends_on} in depends_on that does not exist",
                        )
                else:
                    tt = constants.SECTION_SINGULAR_TO_PLURAL.get(
                        depends_on.get("type", constants.LAUNCH)
                    )
                    dd = depends_on.get("name")
                    if manifest.get(tt).get(dd) is None:
                        print_utils.warn(
                            f"{collection_type}.{collection_name} uses {depends_on} in depends_on that does not exist",
                        )

            #
            # Check depends_on is present when parameters names match outputs defined elsewhere
            #
            for parameter_name, parameter_details in collection_item.get(
                "parameters", {}
            ).items():
                if parameter_details.get("ssm"):
                    output_name = parameter_details.get("ssm").get("name")
                    for (
                        needle_section_name
                    ) in constants.ALL_SECTION_NAMES_THAT_GENERATE_OUTPUTS:
                        for needle_action_name, needle_action_details in manifest.get(
                            needle_section_name, {}
                        ).items():
                            for needle_output in needle_action_details.get(
                                "outputs", {}
                            ).get("ssm", []):
                                if output_name == needle_output.get("param_name"):
                                    found = False
                                    for d in collection_item.get("depends_on", []):
                                        if isinstance(d, str):
                                            dependency = dict(
                                                type=constants.LAUNCH,
                                                name=d,
                                                affinity=constants.LAUNCH,
                                            )
                                        else:
                                            dependency = d
                                        plural = (
                                            constants.SECTION_SINGULAR_TO_PLURAL.get(
                                                dependency.get("type", constants.LAUNCH)
                                            )
                                        )
                                        if (
                                            dependency.get("name") == needle_action_name
                                            and plural == needle_section_name
                                        ):
                                            found = True
                                    if not found:
                                        print_utils.error(
                                            f"{output_name} is used in {collection_type}.{collection_name} from {needle_section_name}.{needle_action_name} but is not in depends_on"
                                        )

            #
            # build graph to check for issues
            #
            uid = f"{collection_type}|{collection_name}"
            data = collection_item
            graph.add_nodes_from(
                [
                    (uid, data),
                ]
            )
            for d in collection_item.get("depends_on", []):
                if isinstance(d, str):
                    dependency = dict(
                        type=constants.LAUNCH, name=d, affinity=constants.LAUNCH
                    )
                else:
                    dependency = d
                plural = constants.SECTION_SINGULAR_TO_PLURAL.get(
                    dependency.get("type", constants.LAUNCH)
                )
                duid = f"{plural}|{dependency.get('name')}"
                graph.add_edge(uid, duid)
    try:
        cycle = nx.find_cycle(graph)
        raise Exception(
            f"found cyclic dependency in your manifest file between: {cycle}"
        )
    except nx.exception.NetworkXNoCycle:
        pass

    print_utils.echo("Finished validating: {}".format(f.name))
    print_utils.echo("Finished validating: OK")


def import_product_set(f, name, portfolio_name):
    url = f"https://raw.githubusercontent.com/awslabs/aws-service-catalog-products/master/{name}/manifest.yaml"
    response = requests.get(url)
    logger.info(f"Getting {url}")
    manifest = serialisation_utils.load(f.read())
    if manifest.get("launches") is None:
        manifest["launches"] = {}
    manifest_segment = serialisation_utils.load(response.text)
    for launch_name, details in manifest_segment.get("launches").items():
        details["portfolio"] = portfolio_name
        manifest["launches"][launch_name] = details
    with open(f.name, "w") as f:
        f.write(serialisation_utils.dump(manifest))


def get_manifest():
    with betterboto_client.ClientContextManager("codecommit") as codecommit:
        content = codecommit.get_file(
            repositoryName=constants.SERVICE_CATALOG_PUPPET_REPO_NAME,
            filePath="manifest.yaml",
        ).get("fileContent")
        return serialisation_utils.load(content)


def save_manifest(manifest):
    with betterboto_client.ClientContextManager("codecommit") as codecommit:
        parent_commit_id = (
            codecommit.get_branch(
                repositoryName=constants.SERVICE_CATALOG_PUPPET_REPO_NAME,
                branchName="master",
            )
            .get("branch")
            .get("commitId")
        )
        codecommit.put_file(
            repositoryName=constants.SERVICE_CATALOG_PUPPET_REPO_NAME,
            branchName="master",
            fileContent=serialisation_utils.dump(manifest),
            parentCommitId=parent_commit_id,
            commitMessage="Auto generated commit",
            filePath=f"manifest.yaml",
        )


def add_to_accounts(account_or_ou):
    manifest = get_manifest()
    manifest.get("accounts").append(account_or_ou)
    save_manifest(manifest)


def remove_from_accounts(account_id_or_ou_id_or_ou_path):
    manifest = get_manifest()
    for account in manifest.get("accounts", []):
        if account.get("account_id", "") == account_id_or_ou_id_or_ou_path:
            manifest.get("accounts").remove(account)
            return save_manifest(manifest)
        elif account.get("ou", "") == account_id_or_ou_id_or_ou_path:
            manifest.get("accounts").remove(account)
            return save_manifest(manifest)
    raise Exception(f"Did not remove {account_id_or_ou_id_or_ou_path}")


def add_to_launches(launch_name, launch):
    manifest = get_manifest()
    launches = manifest.get("launches", {})
    launches[launch_name] = launch
    manifest["launches"] = launches
    save_manifest(manifest)


def remove_from_launches(launch_name):
    manifest = get_manifest()
    del manifest.get("launches")[launch_name]
    save_manifest(manifest)



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

EXECUTOR_ACCOUNT_ID = "SCT_EXECUTOR_ACCOUNT_ID"
DRIFT_TOKEN = "SCT_DRIFT_TOKEN"
RUN_TOKEN = "SCT_RUN_TOKEN"
EXECUTION_MODE = "SCT_EXECUTION_MODE"
PUPPET_ACCOUNT_ID = "SCT_PUPPET_ACCOUNT_ID"
SINGLE_ACCOUNT_ID = "SCT_SINGLE_ACCOUNT_ID"
NUM_WORKERS = "SCT_NUM_WORKERS"
SHOULD_DELETE_ROLLBACK_COMPLETE_STACKS = "SCT_SHOULD_DELETE_ROLLBACK_COMPLETE_STACKS"
SHOULD_USE_PRODUCT_PLANS = "SCT_SHOULD_USE_PRODUCT_PLANS"
INITIALISER_STACK_TAGS = "SCT_INITIALISER_STACK_TAGS"
VERSION = "SCT_VERSION"
HOME_REGION = "SCT_HOME_REGION"
REGIONS = "SCT_REGIONS"
SHOULD_USE_SNS = "SCT_SHOULD_USE_SNS"
SHOULD_FORWARD_EVENTS_TO_EVENTBRIDGE = "SCT_SHOULD_FORWARD_EVENTS_TO_EVENTBRIDGE"
SHOULD_FORWARD_FAILURES_TO_OPSCENTER = "SCT_SHOULD_FORWARD_FAILURES_TO_OPSCENTER"
OUTPUT_CACHE_STARTING_POINT = "SCT_OUTPUT_CACHE_STARTING_POINT"
IS_CACHING_ENABLED = "SCT_IS_CACHING_ENABLED"
GLOBAL_SHARING_MODE = "SCT_GLOBAL_SHARING_MODE"
GLOBAL_SHARE_TAG_OPTIONS = "SCT_GLOBAL_SHARE_TAG_OPTIONS"
ON_COMPLETE_URL = "SCT_ON_COMPLETE_URL"
SPOKE_EXECUTION_MODE_DEPLOY_ENV = "SCT_SPOKE_EXECUTION_MODE_DEPLOY_ENV"
SCHEDULER_THREADS_OR_PROCESSES = "SCT_SCHEDULER_THREADS_OR_PROCESSES"
SCHEDULER_ALGORITHM = "SCT_SCHEDULER_ALGORITHM"
SPOKE_SCHEDULER_THREADS_OR_PROCESSES = "SCT_SPOKE_SCHEDULER_THREADS_OR_PROCESSES"
SPOKE_SCHEDULER_ALGORITHM = "SCT_SPOKE_SCHEDULER_ALGORITHM"
GLOBAL_SHARE_PRINCIPALS = "SCT_GLOBAL_SHARE_PRINCIPALS"
AWS_STS_REGIONAL_ENDPOINTS = "AWS_STS_REGIONAL_ENDPOINTS"



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import json

import cfn_tools
from botocore.exceptions import ClientError

import servicecatalog_puppet.manifest_utils
import servicecatalog_puppet.serialisation_utils
from servicecatalog_puppet import constants
from servicecatalog_puppet.workflow.stack import provision_stack_task


class ProvisionStackDryRunTask(provision_stack_task.ProvisionStackTask):
    def write_result(
        self, current_version, new_version, effect, current_status, active, notes=""
    ):
        with self.output().open("w") as f:
            f.write(
                json.dumps(
                    {
                        "current_version": current_version,
                        "new_version": new_version,
                        "effect": effect,
                        "current_status": current_status,
                        "active": active,
                        "notes": notes,
                        "params": self.param_kwargs,
                    },
                    indent=4,
                    default=str,
                )
            )

    def get_current_status(self):
        with self.spoke_regional_client("cloudformation") as cloudformation:
            try:
                paginator = cloudformation.get_paginator("describe_stacks")
                for page in paginator.paginate(
                    StackName=self.stack_name,
                ):
                    for stack in page.get("Stacks", []):
                        status = stack.get("StackStatus")
                        if status != "DELETE_COMPLETE":
                            return status
            except ClientError as error:
                if (
                    error.response["Error"]["Message"]
                    != f"Stack with id {self.stack_name} does not exist"
                ):
                    raise error
        return "-"

    def run(self):
        status = self.get_current_status()

        if status == "-":
            self.write_result(
                "-",
                self.version_id,
                effect=constants.CHANGE,
                current_status="-",
                active="N/A",
                notes="Stack would be created",
            )
        elif status == "ROLLBACK_COMPLETE":
            if self.should_delete_rollback_complete_stacks:
                self.write_result(
                    "-",
                    self.version_id,
                    effect=constants.CHANGE,
                    current_status="-",
                    active="N/A",
                    notes="Stack would be replaced",
                )
            else:
                self.write_result(
                    "-",
                    "-",
                    effect=constants.NO_CHANGE,
                    current_status="-",
                    active="N/A",
                    notes="Stack needs remediation - it's in ROLLBACK_COMPLETE",
                )
        else:
            task_output = dict(
                **self.params_for_results_display(),
                account_parameters=servicecatalog_puppet.serialisation_utils.unwrap(
                    self.account_parameters
                ),
                launch_parameters=servicecatalog_puppet.serialisation_utils.unwrap(
                    self.launch_parameters
                ),
                manifest_parameters=servicecatalog_puppet.serialisation_utils.unwrap(
                    self.manifest_parameters
                ),
            )

            all_params = self.get_parameter_values()

            template_to_provision_source = self.input().get("template").open("r").read()
            try:
                template_to_provision = cfn_tools.load_yaml(
                    template_to_provision_source
                )
            except Exception:
                try:
                    template_to_provision = cfn_tools.load_json(
                        template_to_provision_source
                    )
                except Exception:
                    raise Exception("Could not parse new template as YAML or JSON")

            params_to_use = dict()
            for param_name, p in template_to_provision.get("Parameters", {}).items():
                if all_params.get(param_name, p.get("DefaultValue")) is not None:
                    params_to_use[param_name] = all_params.get(
                        param_name, p.get("DefaultValue")
                    )

            existing_stack_params_dict = dict()
            existing_template = ""
            if status in [
                "CREATE_COMPLETE",
                "UPDATE_ROLLBACK_COMPLETE",
                "UPDATE_COMPLETE",
                "IMPORT_COMPLETE",
                "IMPORT_ROLLBACK_COMPLETE",
            ]:
                with self.spoke_regional_client("cloudformation") as cloudformation:
                    existing_stack_params_dict = {}
                    stack = cloudformation.describe_stacks(
                        StackName=self.stack_name
                    ).get("Stacks")[0]
                    summary_response = cloudformation.get_template_summary(
                        StackName=self.stack_name,
                    )
                    for parameter in summary_response.get("Parameters"):
                        existing_stack_params_dict[parameter.get("ParameterKey")] = (
                            parameter.get("DefaultValue")
                        )
                    for stack_param in stack.get("Parameters", []):
                        existing_stack_params_dict[stack_param.get("ParameterKey")] = (
                            stack_param.get("ParameterValue")
                        )
                    template_body = cloudformation.get_template(
                        StackName=self.stack_name, TemplateStage="Original"
                    ).get("TemplateBody")
                    try:
                        existing_template = cfn_tools.load_yaml(template_body)
                    except Exception:
                        try:
                            existing_template = cfn_tools.load_json(template_body)
                        except Exception:
                            raise Exception(
                                "Could not parse existing template as YAML or JSON"
                            )

            template_to_use = cfn_tools.dump_yaml(template_to_provision)
            if status == "UPDATE_ROLLBACK_COMPLETE":
                self.write_result(
                    "?",
                    self.version_id,
                    effect=constants.CHANGE,
                    current_status=status,
                    active="N/A",
                    notes="Stack would be updated",
                )
            else:
                if existing_stack_params_dict == params_to_use:
                    self.info(f"params unchanged")
                    if template_to_use == cfn_tools.dump_yaml(existing_template):
                        self.info(f"template the same")
                        self.write_result(
                            "?",
                            self.version_id,
                            effect=constants.NO_CHANGE,
                            current_status=status,
                            active="N/A",
                            notes="No change",
                        )
                    else:
                        self.info(f"template changed")
                        self.write_result(
                            "?",
                            self.version_id,
                            effect=constants.CHANGE,
                            current_status=status,
                            active="N/A",
                            notes="Template has changed",
                        )
                else:
                    self.info(f"params changed")
                    self.write_result(
                        "?",
                        self.version_id,
                        effect=constants.CHANGE,
                        current_status=status,
                        active="N/A",
                        notes="Parameters have changed",
                    )



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import luigi

from servicecatalog_puppet import constants
from servicecatalog_puppet.workflow.dependencies import tasks


class DoExecuteCodeBuildRunTask(tasks.TaskWithParameters):
    code_build_run_name = luigi.Parameter()

    region = luigi.Parameter()
    account_id = luigi.Parameter()

    project_name = luigi.Parameter()

    manifest_file_path = luigi.Parameter()

    section_name = constants.CODE_BUILD_RUNS
    cachable_level = constants.CACHE_LEVEL_RUN

    @property
    def item_name(self):
        return self.code_build_run_name

    def params_for_results_display(self):
        return {
            "puppet_account_id": self.puppet_account_id,
            "code_build_run_name": self.code_build_run_name,
            "region": self.region,
            "account_id": self.account_id,
        }

    def run(self):
        with self.hub_client("codebuild") as codebuild:
            provided_parameters = self.get_parameter_values()
            parameters_to_use = list()

            projects = codebuild.batch_get_projects(names=[self.project_name]).get(
                "projects", []
            )
            for project in projects:
                if project.get("name") == self.project_name:
                    for environment_variable in project.get("environment", {}).get(
                        "environmentVariables", []
                    ):
                        if environment_variable.get("type") == "PLAINTEXT":
                            n = environment_variable.get("name")
                            if provided_parameters.get(n):
                                parameters_to_use.append(
                                    dict(
                                        name=n,
                                        value=provided_parameters.get(n),
                                        type="PLAINTEXT",
                                    )
                                )

            parameters_to_use.append(
                dict(
                    name="TARGET_ACCOUNT_ID",
                    value=self.account_id,
                    type="PLAINTEXT",
                )
            )
            parameters_to_use.append(
                dict(
                    name="TARGET_REGION",
                    value=self.region,
                    type="PLAINTEXT",
                )
            )
            result = codebuild.start_build_and_wait_for_completion(
                projectName=self.project_name,
                environmentVariablesOverride=parameters_to_use,
            )
            if result.get("buildStatus") != "SUCCEEDED":
                raise Exception(
                    f"Executing CodeBuild failed: {result.get('buildStatus')}"
                )
        self.write_empty_output()



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0
import time

import luigi

from servicecatalog_puppet import config, constants
from servicecatalog_puppet.workflow.dependencies import tasks


class CreateAssociationTask(tasks.TaskWithReference):
    account_id = luigi.Parameter()
    region = luigi.Parameter()
    portfolio = luigi.Parameter()
    portfolio_task_reference = luigi.Parameter()

    cachable_level = constants.CACHE_LEVEL_RUN

    def params_for_results_display(self):
        return {
            "task_reference": self.task_reference,
            "puppet_account_id": self.puppet_account_id,
            "portfolio": self.portfolio,
            "region": self.region,
            "account_id": self.account_id,
        }

    def run(self):
        portfolio_details = self.get_output_from_reference_dependency(
            self.portfolio_task_reference
        )
        portfolio_id = portfolio_details.get("Id")

        principal_to_associate = config.get_puppet_role_arn(self.account_id)
        was_present = self.check_was_present(portfolio_id, principal_to_associate)
        self.info(f"Check for association: {was_present}")
        if not was_present:
            with self.spoke_regional_client("servicecatalog") as servicecatalog:
                servicecatalog.associate_principal_with_portfolio(
                    PortfolioId=portfolio_id,
                    PrincipalARN=principal_to_associate,
                    PrincipalType="IAM",
                )
        while not was_present:
            time.sleep(1)
            was_present = self.check_was_present(portfolio_id, principal_to_associate)
            self.info(f"Check for association: {was_present}")

        self.write_empty_output()

    def check_was_present(self, portfolio_id, principal_to_associate):
        with self.spoke_regional_client("servicecatalog") as servicecatalog:
            paginator = servicecatalog.get_paginator("list_principals_for_portfolio")
            for page in paginator.paginate(PortfolioId=portfolio_id):
                for principal in page.get("Principals", []):
                    if principal_to_associate == principal.get("PrincipalARN"):
                        return True
        return False



#   Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#   SPDX-License-Identifier: Apache-2.0

from servicecatalog_puppet import constants, task_reference_constants


def generate_task_reference(
    parameter_details,
    account_id_to_use_for_boto3_call,
    region_to_use_for_boto3_call,
):
    if parameter_details.get("cloudformation_stack_output"):
        return (
            f"{constants.BOTO3_PARAMETERS}"
            f"-cloudformation_stack_output"
            f"-{parameter_details.get('cloudformation_stack_output').get('stack_name')}"
            f"-{parameter_details.get('cloudformation_stack_output').get('output_key')}"
            f"-{account_id_to_use_for_boto3_call}"
            f"-{region_to_use_for_boto3_call}"
        )
    elif parameter_details.get("servicecatalog_provisioned_product_output"):
        return (
            f"{constants.BOTO3_PARAMETERS}"
            f"-servicecatalog_provisioned_product_output"
            f"-{parameter_details.get('servicecatalog_provisioned_product_output').get('provisioned_product_name')}"
            f"-{parameter_details.get('servicecatalog_provisioned_product_output').get('output_key')}"
            f"-{account_id_to_use_for_boto3_call}"
            f"-{region_to_use_for_boto3_call}"
        )

    elif parameter_details.get("boto3"):
        return (
            f"{constants.BOTO3_PARAMETERS}"
            f"-{task.get('section_name')}"
            f"-{task.get('item_name')}"
            f"-{parameter_name}"
            f"-{account_id_to_use_for_boto3_call}"
            f"-{region_to_use_for_boto3_call}"
        )


def boto3_parameter_handler(
    new_tasks, parameter_details, parameter_name, puppet_account_id, task
):
    if parameter_details.get("boto3"):
        boto3_parameter_details = parameter_details.get("boto3")
        account_id_to_use_for_boto3_call = (
            str(boto3_parameter_details.get("account_id", puppet_account_id))
            .replace("${AWS::AccountId}", task.get("account_id"))
            .replace("${AWS::PuppetAccountId}", puppet_account_id)
        )
        region_to_use_for_boto3_call = boto3_parameter_details.get(
            "region", constants.HOME_REGION
        ).replace("${AWS::Region}", task.get("region"))

        boto3_parameter_task_reference = generate_task_reference(
            parameter_details,
            account_id_to_use_for_boto3_call,
            region_to_use_for_boto3_call,
        )

        task_execution = task.get("execution", constants.EXECUTION_MODE_DEFAULT)
        if task.get(task_execution) in [
            constants.EXECUTION_MODE_HUB,
            constants.EXECUTION_MODE_ASYNC,
        ]:
            if account_id_to_use_for_boto3_call != puppet_account_id:
                raise Exception(
                    f"Cannot use cross account Boto3 Parameters in execution mode: {task_execution}"
                )
        if not new_tasks.get(boto3_parameter_task_reference):
            new_tasks[boto3_parameter_task_reference] = {
                "status": task.get("status"),
                "execution": task_execution,
                "task_reference": boto3_parameter_task_reference,
                "dependencies_by_reference": [],
                "dependencies": [],
                task_reference_constants.MANIFEST_SECTION_NAMES: dict(),
                task_reference_constants.MANIFEST_ITEM_NAMES: dict(),
                task_reference_constants.MANIFEST_ACCOUNT_IDS: dict(),
                "account_id": account_id_to_use_for_boto3_call,
                "region": region_to_use_for_boto3_call,
                "arguments": boto3_parameter_details.get("arguments"),
                "call": boto3_parameter_details.get("call"),
                "client": boto3_parameter_details.get("client"),
                "filter": boto3_parameter_details.get("filter"),
                "use_paginator": boto3_parameter_details.get("use_paginator"),
                "section_name": constants.BOTO3_PARAMETERS,
            }

        boto3_task = new_tasks[boto3_parameter_task_reference]
        boto3_task[task_reference_constants.MANIFEST_SECTION_NAMES].update(
            task.get(task_reference_constants.MANIFEST_SECTION_NAMES)
        )
        boto3_task[task_reference_constants.MANIFEST_ITEM_NAMES].update(
            task.get(task_reference_constants.MANIFEST_ITEM_NAMES)
        )
        boto3_task[task_reference_constants.MANIFEST_ACCOUNT_IDS].update(
            task.get(task_reference_constants.MANIFEST_ACCOUNT_IDS)
        )

        task["dependencies_by_reference"].append(boto3_parameter_task_reference)

        if not task.get("boto3_parameters_tasks_references"):
            task["boto3_parameters_tasks_references"] = dict()

        task["boto3_parameters_tasks_references"][
            parameter_name
        ] = boto3_parameter_task_reference



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import time

import luigi
import yaml

from servicecatalog_puppet import constants
from servicecatalog_puppet.workflow.dependencies import tasks


class SharePortfolioViaOrgsTask(tasks.TaskWithReference):
    region = luigi.Parameter()
    portfolio = luigi.Parameter()
    share_tag_options = luigi.BoolParameter()
    share_principals = luigi.BoolParameter()
    ou_to_share_with = luigi.Parameter()
    portfolio_task_reference = luigi.Parameter()
    describe_portfolio_shares_task_ref = luigi.Parameter()
    cachable_level = constants.CACHE_LEVEL_RUN

    def params_for_results_display(self):
        return {
            "puppet_account_id": self.puppet_account_id,
            "portfolio": self.portfolio,
            "region": self.region,
            "ou_to_share_with": self.ou_to_share_with,
            "share_tag_options": self.share_tag_options,
            "share_principals": self.share_principals,
        }

    def run(self):
        portfolio_id = self.get_attribute_from_output_from_reference_dependency(
            "Id", self.portfolio_task_reference
        )

        changes = dict()
        existing_share_details = (
            self.get_attribute_from_output_from_reference_dependency(
                self.ou_to_share_with, self.describe_portfolio_shares_task_ref
            )
        )

        if existing_share_details:
            if existing_share_details.get("ShareTagOptions") != self.share_tag_options:
                changes["ShareTagOptions"] = self.share_tag_options
            if existing_share_details.get("SharePrincipals") != self.share_principals:
                changes["SharePrincipals"] = self.share_principals

            if changes:
                with self.hub_regional_client("servicecatalog") as servicecatalog:
                    servicecatalog.update_portfolio_share(
                        PortfolioId=portfolio_id,
                        OrganizationNode={
                            "Type": existing_share_details.get("Type"),
                            "Value": self.ou_to_share_with,
                        },
                        **changes,
                    )
        else:
            with self.hub_regional_client("servicecatalog") as servicecatalog:
                portfolio_share_token = servicecatalog.create_portfolio_share(
                    PortfolioId=portfolio_id,
                    ShareTagOptions=self.share_tag_options,
                    SharePrincipals=self.share_principals,
                    OrganizationNode=dict(
                        Type="ORGANIZATIONAL_UNIT", Value=self.ou_to_share_with
                    ),
                ).get("PortfolioShareToken")

                status = "NOT_STARTED"

                while status in ["NOT_STARTED", "IN_PROGRESS"]:
                    time.sleep(5)
                    response = servicecatalog.describe_portfolio_share_status(
                        PortfolioShareToken=portfolio_share_token
                    )
                    status = response.get("Status")
                    self.info(f"New status: {status}")

                if status in ["COMPLETED_WITH_ERRORS", "ERROR"]:
                    errors = list()
                    for error in response.get("ShareDetails").get("ShareErrors"):
                        if error.get("Error") == "DuplicateResourceException":
                            self.warning(yaml.safe_dump(error))
                        else:
                            errors.append(error)
                    if len(errors) > 0:
                        raise Exception(yaml.safe_dump(response.get("ShareDetails")))

        self.write_empty_output()



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0
import time

import luigi

from servicecatalog_puppet import constants
from servicecatalog_puppet.workflow.dependencies import tasks


class ProvisioningArtifactParametersTask(tasks.TaskWithReference):
    portfolio = luigi.Parameter()
    product = luigi.Parameter()
    version = luigi.Parameter()
    region = luigi.Parameter()
    cachable_level = constants.CACHE_LEVEL_RUN

    @property
    def retry_count(self):
        return 5

    def params_for_results_display(self):
        return {
            "puppet_account_id": self.puppet_account_id,
            "portfolio": self.portfolio,
            "region": self.region,
            "product": self.product,
            "version": self.version,
        }

    def run(self):
        with self.hub_regional_client("servicecatalog") as service_catalog:
            provisioning_artifact_parameters = None
            retries = 3
            while retries > 0:
                try:
                    provisioning_artifact_parameters = (
                        service_catalog.describe_provisioning_parameters(
                            ProductName=self.product,
                            ProvisioningArtifactName=self.version,
                            PathName=self.portfolio,
                        ).get("ProvisioningArtifactParameters", [])
                    )
                    retries = 0
                    break
                except service_catalog.exceptions.ClientError as ex:
                    if "S3 error: Access Denied" in str(ex):
                        self.info("Swallowing S3 error: Access Denied")
                    else:
                        raise ex
                    time.sleep(3)
                    retries -= 1

            self.write_output(
                provisioning_artifact_parameters
                if isinstance(provisioning_artifact_parameters, list)
                else [provisioning_artifact_parameters]
            )



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0
import os

import botocore
import luigi

from servicecatalog_puppet import constants
from servicecatalog_puppet.workflow.dependencies import tasks


class GetCloudFormationTemplateFromS3(tasks.TaskWithReference):
    account_id = luigi.Parameter()

    bucket = luigi.Parameter()
    key = luigi.Parameter()
    region = luigi.Parameter()
    version_id = luigi.Parameter()
    cachable_level = constants.CACHE_LEVEL_NO_CACHE

    def params_for_results_display(self):
        return {
            "task_reference": self.task_reference,
            "bucket": self.bucket,
            "key": self.key.replace("-${AWS::Region}", f"-{self.region}"),
            "region": self.region,
            "version_id": self.version_id,
        }

    def run(self):
        with self.hub_client("s3") as s3:
            regional_template = self.key.replace("-${AWS::Region}", f"-{self.region}")
            global_template = self.key.replace("-${AWS::Region}", "")
            p = dict(Bucket=self.bucket)
            if self.version_id != "":
                p["VersionId"] = self.version_id

            output = self.output_location_non_cached
            self.info(f"Trying regional template: {regional_template} to {output}")
            if not os.path.exists(os.path.dirname(output)):
                os.makedirs(os.path.dirname(output))

            try:
                s3.download_file(Key=regional_template, Filename=output, **p)
            except botocore.exceptions.ClientError as e:
                if "404" == str(e.response["Error"]["Code"]):
                    self.info(
                        f"Didnt find regional template: {regional_template}, will try global: {global_template}"
                    )
                    s3.download_file(Key=global_template, Filename=output, **p)
                else:
                    raise e



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

from servicecatalog_puppet import constants, task_reference_constants


def manifest_related_args(task_to_add):
    return {
        task_reference_constants.MANIFEST_SECTION_NAMES: dict(
            **task_to_add.get(task_reference_constants.MANIFEST_SECTION_NAMES)
        ),
        task_reference_constants.MANIFEST_ITEM_NAMES: dict(
            **task_to_add.get(task_reference_constants.MANIFEST_ITEM_NAMES)
        ),
        task_reference_constants.MANIFEST_ACCOUNT_IDS: dict(
            **task_to_add.get(task_reference_constants.MANIFEST_ACCOUNT_IDS)
        ),
    }


def create_custodian_role(
    task_to_add,
    all_tasks,
    account_id,
    custodian_account_id,
    custodian_region,
    custodian_role_name,
    custodian_role_path,
    custodian_role_managed_policy_arns,
):
    create_custodian_role_ref = (
        f"{constants.C7N_CREATE_CUSTODIAN_ROLE_TASK}-{account_id}"
    )
    all_tasks[create_custodian_role_ref] = dict(
        section_name=constants.C7N_CREATE_CUSTODIAN_ROLE_TASK,
        task_reference=create_custodian_role_ref,
        account_id=account_id,
        region=custodian_region,
        c7n_account_id=custodian_account_id,
        role_name=custodian_role_name,
        role_path=custodian_role_path,
        role_managed_policy_arns=custodian_role_managed_policy_arns,
        dependencies_by_reference=[],
        execution=task_to_add.get("execution"),
        **manifest_related_args(task_to_add),
    )


def prepare_account_as_hub(
    task_to_add,
    all_tasks,
    custodian_account_id,
    custodian_region,
    custodian_role_name,
    custodian_role_path,
    schedule_expression,
    custodian_c7n_version,
    custodian_c7n_org_version,
    organization,
):
    create_custodian_event_bus_task_ref = (
        f"{constants.C7N_PREPARE_HUB_ACCOUNT_TASK}-{custodian_account_id}"
    )
    all_tasks[create_custodian_event_bus_task_ref] = dict(
        task_reference=create_custodian_event_bus_task_ref,
        section_name=constants.C7N_PREPARE_HUB_ACCOUNT_TASK,
        account_id=custodian_account_id,
        region=custodian_region,
        custodian_region=custodian_region,
        c7n_version=custodian_c7n_version,
        c7n_org_version=custodian_c7n_org_version,
        organization=organization,
        role_name=custodian_role_name,
        role_path=custodian_role_path,
        schedule_expression=schedule_expression,
        dependencies_by_reference=[],
        execution=task_to_add.get("execution"),
        **manifest_related_args(task_to_add),
    )


def forward_events_account_task(
    task_to_add, all_tasks, account_id, custodian_account_id, custodian_region
):
    forward_events_for_account_tasks = (
        f"{constants.C7N_FORWARD_EVENTS_FOR_ACCOUNT_TASK}-{account_id}"
    )
    all_tasks[forward_events_for_account_tasks] = dict(
        section_name=constants.C7N_FORWARD_EVENTS_FOR_ACCOUNT_TASK,
        task_reference=forward_events_for_account_tasks,
        account_id=account_id,
        region=custodian_region,
        c7n_account_id=custodian_account_id,
        custodian_region=custodian_region,
        dependencies_by_reference=[
            f"{constants.C7N_PREPARE_HUB_ACCOUNT_TASK}-{custodian_account_id}",
        ],
        execution=task_to_add.get("execution"),
        **manifest_related_args(task_to_add),
    )


def forward_events_region_task(
    task_to_add, all_tasks, account_id, region, custodian_account_id, custodian_region
):
    if account_id == custodian_account_id and region == custodian_region:
        return

    forward_events_for_region_task = (
        f"{constants.C7N_FORWARD_EVENTS_FOR_REGION_TASK}-{account_id}-{region}"
    )
    all_tasks[forward_events_for_region_task] = dict(
        section_name=constants.C7N_FORWARD_EVENTS_FOR_REGION_TASK,
        task_reference=forward_events_for_region_task,
        account_id=account_id,
        region=region,
        c7n_account_id=custodian_account_id,
        custodian_region=custodian_region,
        dependencies_by_reference=[
            f"{constants.C7N_FORWARD_EVENTS_FOR_ACCOUNT_TASK}-{account_id}",
        ],
        execution=task_to_add.get("execution"),
        **manifest_related_args(task_to_add),
    )


def handle_deploy_policies(
    task_to_add,
    all_tasks,
    account_id,
    region,
    custodian_account_id,
    custodian_region,
    role_name,
    role_path,
    uses_orgs,
):
    deploy_policies_task_ref = (
        f"{constants.C7N_DEPLOY_POLICIES_TASK}-{custodian_account_id}"
    )
    if all_tasks.get(deploy_policies_task_ref) is None:
        all_tasks[deploy_policies_task_ref] = dict(
            section_name=constants.C7N_DEPLOY_POLICIES_TASK,
            task_reference=deploy_policies_task_ref,
            account_id=custodian_account_id,
            region=custodian_region,
            policies=task_to_add.get("policies"),
            deployments=dict(),
            role_name=role_name,
            role_path=role_path,
            uses_orgs=uses_orgs,
            dependencies_by_reference=[
                # custodian dependencies
                f"{constants.C7N_CREATE_CUSTODIAN_ROLE_TASK}-{custodian_account_id}",
                f"{constants.C7N_PREPARE_HUB_ACCOUNT_TASK}-{custodian_account_id}",
            ],
            execution=task_to_add.get("execution"),
            **manifest_related_args(task_to_add),
        )
    deploy_policies_task = all_tasks[deploy_policies_task_ref]
    if deploy_policies_task["deployments"].get(region) is None:
        deploy_policies_task["deployments"][region] = list()
    deploy_policies_task["deployments"][region].append(account_id)


def handle_c7n_aws_lambdas_for_custodian(
    task_to_add,
    all_tasks,
    region,
    custodian_account_id,
    custodian_region,
    custodian_role_name,
    custodian_role_path,
    schedule_expression,
    custodian_role_managed_policy_arns,
    custodian_c7n_version,
    custodian_c7n_org_version,
    organization,
):
    create_custodian_role(
        task_to_add,
        all_tasks,
        custodian_account_id,
        custodian_account_id,
        custodian_region,
        custodian_role_name,
        custodian_role_path,
        custodian_role_managed_policy_arns,
    )
    prepare_account_as_hub(
        task_to_add,
        all_tasks,
        custodian_account_id,
        custodian_region,
        custodian_role_name,
        custodian_role_path,
        schedule_expression,
        custodian_c7n_version,
        custodian_c7n_org_version,
        organization,
    )
    forward_events_account_task(
        task_to_add,
        all_tasks,
        custodian_account_id,
        custodian_account_id,
        custodian_region,
    )
    forward_events_region_task(
        task_to_add,
        all_tasks,
        custodian_account_id,
        region,
        custodian_account_id,
        custodian_region,
    )


def handle_c7n_aws_lambdas_for_spoke(
    task_to_add,
    all_tasks,
    account_id,
    region,
    custodian_account_id,
    custodian_region,
    custodian_role_name,
    custodian_role_path,
    custodian_role_managed_policy_arns,
):
    create_custodian_role(
        task_to_add,
        all_tasks,
        account_id,
        custodian_account_id,
        custodian_region,
        custodian_role_name,
        custodian_role_path,
        custodian_role_managed_policy_arns,
    )
    forward_events_account_task(
        task_to_add, all_tasks, account_id, custodian_account_id, custodian_region
    )
    forward_events_region_task(
        task_to_add,
        all_tasks,
        account_id,
        region,
        custodian_account_id,
        custodian_region,
    )


def get_organization_for_account(manifest, account_id):
    for a in manifest.get("accounts"):
        if str(a.get("account_id")) == str(account_id):
            return a["organization"]


def get_custodian_region(manifest, account_id):
    for a in manifest.get("accounts"):
        if str(a.get("account_id")) == str(account_id):
            return a["default_region"]


def handle_c7n_aws_lambdas(
    all_tasks,
    all_tasks_task_reference,
    item_name,
    puppet_account_id,
    section_name,
    task_reference,
    task_to_add,
    manifest,
):
    account_id = task_to_add.get("account_id")
    region = task_to_add.get("region")
    custodian_account_id = task_to_add.get("custodian")
    custodian_role_name = task_to_add.get("role_name")
    custodian_role_path = task_to_add.get("role_path")
    schedule_expression = task_to_add.get("schedule_expression", "")
    custodian_role_managed_policy_arns = task_to_add.get("role_managed_policy_arns")
    custodian_c7n_version = task_to_add.get("c7n_version")
    custodian_c7n_org_version = task_to_add.get("c7n_org_version")
    organization = get_organization_for_account(manifest, account_id)
    custodian_region = get_custodian_region(manifest, custodian_account_id)

    handle_c7n_aws_lambdas_for_custodian(
        task_to_add,
        all_tasks,
        region,
        custodian_account_id,
        custodian_region,
        custodian_role_name,
        custodian_role_path,
        schedule_expression,
        custodian_role_managed_policy_arns,
        custodian_c7n_version,
        custodian_c7n_org_version,
        organization,
    )
    handle_c7n_aws_lambdas_for_spoke(
        task_to_add,
        all_tasks,
        account_id,
        region,
        custodian_account_id,
        custodian_region,
        custodian_role_name,
        custodian_role_path,
        custodian_role_managed_policy_arns,
    )

    handle_deploy_policies(
        task_to_add,
        all_tasks,
        account_id,
        region,
        custodian_account_id,
        custodian_region,
        custodian_role_name,
        custodian_role_path,
        uses_orgs=custodian_c7n_org_version != "",
    )

    del all_tasks[all_tasks_task_reference]



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0
import json
from copy import deepcopy

import jmespath
import luigi

from servicecatalog_puppet import constants
from servicecatalog_puppet.workflow.dependencies import tasks


class GetSSMParameterTask(tasks.TaskWithReference):
    account_id = luigi.Parameter()
    param_name = luigi.Parameter()
    jmespath_location = luigi.Parameter()
    region = luigi.Parameter()
    cachable_level = constants.CACHE_LEVEL_RUN

    def params_for_results_display(self):
        return {
            "task_reference": self.task_reference,
            "account_id": self.account_id,
            "region": self.region,
            "param_name": self.param_name,
            "jmespath_location": self.jmespath_location,
        }

    def get_parameter_name_to_use(self):
        return self.param_name.replace("${AWS::Region}", self.region).replace(
            "${AWS::AccountId}", self.account_id
        )

    def run(self):
        parameter_name_to_use = self.get_parameter_name_to_use()
        result = {}
        with self.spoke_regional_client("ssm") as ssm:
            try:
                parameter = ssm.get_parameter(Name=parameter_name_to_use)
                p = parameter.get("Parameter")
                if self.jmespath_location:
                    p = deepcopy(parameter.get("Parameter"))
                    p["Value"] = jmespath.search(
                        self.jmespath_location, json.loads(p.get("Value"))
                    )
                result = {parameter_name_to_use: p}
            except ssm.exceptions.ParameterNotFound:
                pass

        self.write_output(result)


class GetSSMParameterByPathTask(tasks.TaskWithReference):
    account_id = luigi.Parameter()
    path = luigi.Parameter()
    jmespath_location = luigi.Parameter()
    region = luigi.Parameter()
    cachable_level = constants.CACHE_LEVEL_RUN

    def params_for_results_display(self):
        return {
            "task_reference": self.task_reference,
            "account_id": self.account_id,
            "region": self.region,
            "path": self.path,
            "jmespath_location": self.jmespath_location,
        }

    def run(self):
        parameters = dict()
        with self.spoke_regional_client("ssm") as ssm:
            paginator = ssm.get_paginator("get_parameters_by_path")
            for page in paginator.paginate(
                Path=self.get_parameter_path_to_use(), Recursive=True
            ):
                for parameter in page.get("Parameters", []):
                    p = deepcopy(parameter)
                    if self.jmespath_location:
                        p["Value"] = jmespath.search(
                            self.jmespath_location, json.loads(p.get("Value"))
                        )
                    parameters[parameter.get("Name")] = p
        self.write_output(parameters)

    def get_parameter_path_to_use(self):
        return self.path.replace("${AWS::Region}", self.region).replace(
            "${AWS::AccountId}", self.account_id
        )



#   Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#   SPDX-License-Identifier: Apache-2.0

from servicecatalog_puppet import constants, task_reference_constants


def assertCrossAccountAccessWillWork(
    owning_account, task, task_execution, puppet_account_id
):
    if task_execution not in [
        constants.EXECUTION_MODE_HUB,
        constants.EXECUTION_MODE_ASYNC,
    ]:
        if (
            owning_account != task.get("account_id")
            and owning_account != puppet_account_id
        ):
            message = f"Cannot use cross account SSM parameters in execution mode: {task_execution}."
            message += f"For task {task.get('task_reference')}, "
            message += f"parameter is in account {owning_account} and task will execute in {task.get('account_id')}."
            raise Exception(message)


def ssm_parameter_handler(
    all_tasks, home_region, new_tasks, parameter_details, puppet_account_id, task
):
    if parameter_details.get("ssm"):
        ssm_parameter_details = parameter_details.get("ssm")
        task_account_id = task.get("account_id")
        task_region = task.get("region")
        parameter_account_id = (
            str(ssm_parameter_details.get("account_id", puppet_account_id))
            .replace("${AWS::AccountId}", task_account_id)
            .replace("${AWS::PuppetAccountId}", puppet_account_id)
        )
        parameter_region = ssm_parameter_details.get("region", home_region).replace(
            "${AWS::Region}", task_region
        )
        ssm_parameter_partial_task_reference = (
            f"{parameter_account_id}-{parameter_region}"
        )
        parameter_name = (
            str(ssm_parameter_details.get("name"))
            .replace("${AWS::Region}", task_region)
            .replace("${AWS::AccountId}", task_account_id)
            .replace("${AWS::PuppetAccountId}", puppet_account_id)
        )

        path = ssm_parameter_details.get("path")
        if path is None:
            parameter_task_reference = f"{constants.SSM_PARAMETERS}-{ssm_parameter_partial_task_reference}-{parameter_name}"
        else:
            parameter_task_reference = f"{constants.SSM_PARAMETERS_WITH_A_PATH}-{ssm_parameter_partial_task_reference}-{path}"

        if all_tasks.get(parameter_task_reference):
            ssm_task_params = all_tasks.get(parameter_task_reference)
        else:
            if parameter_account_id not in [task_account_id, puppet_account_id]:
                raise Exception(
                    f"SSM Parameters can only come from the target account or the puppet account. "
                    + f"{task.get('task_reference')} in {task_account_id} is using {parameter_name} from {parameter_account_id}, "
                    + f"which is not the puppet_account_id {puppet_account_id}"
                )

            parameter_task_execution = task.get(
                "execution", constants.EXECUTION_MODE_DEFAULT
            )

            if (
                parameter_account_id != task_account_id
                and parameter_task_execution == constants.EXECUTION_MODE_SPOKE
            ):
                parameter_task_execution = constants.EXECUTION_MODE_HUB

            ssm_task_params = {
                "task_reference": parameter_task_reference,
                "account_id": parameter_account_id,
                "region": parameter_region,
                "jmespath": ssm_parameter_details.get("jmespath", None),
                task_reference_constants.MANIFEST_SECTION_NAMES: dict(),
                task_reference_constants.MANIFEST_ITEM_NAMES: dict(),
                task_reference_constants.MANIFEST_ACCOUNT_IDS: dict(),
                "dependencies": [],
                "dependencies_by_reference": [],
                "execution": parameter_task_execution,
            }

            if path is None:
                ssm_task_params["param_name"] = parameter_name
                ssm_task_params["section_name"] = constants.SSM_PARAMETERS
            else:
                ssm_task_params["path"] = path
                ssm_task_params["section_name"] = constants.SSM_PARAMETERS_WITH_A_PATH

            new_tasks[parameter_task_reference] = ssm_task_params

        ssm_task_params[task_reference_constants.MANIFEST_SECTION_NAMES].update(
            **task.get(task_reference_constants.MANIFEST_SECTION_NAMES)
        )
        ssm_task_params[task_reference_constants.MANIFEST_ITEM_NAMES].update(
            **task.get(task_reference_constants.MANIFEST_ITEM_NAMES)
        )
        ssm_task_params[task_reference_constants.MANIFEST_ACCOUNT_IDS].update(
            **task.get(task_reference_constants.MANIFEST_ACCOUNT_IDS)
        )

        task["dependencies_by_reference"].append(parameter_task_reference)

        if not task.get("ssm_parameters_tasks_references"):
            task["ssm_parameters_tasks_references"] = dict()

        task["ssm_parameters_tasks_references"][
            parameter_name
        ] = parameter_task_reference



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import json

import luigi

from servicecatalog_puppet import config, constants
from servicecatalog_puppet.workflow.dependencies import tasks


class DoInvokeLambdaTask(tasks.TaskWithParameters):
    lambda_invocation_name = luigi.Parameter()
    region = luigi.Parameter()
    account_id = luigi.Parameter()

    function_name = luigi.Parameter()
    qualifier = luigi.Parameter()
    invocation_type = luigi.Parameter()

    manifest_file_path = luigi.Parameter()

    section_name = constants.LAMBDA_INVOCATIONS
    cachable_level = constants.CACHE_LEVEL_RUN

    @property
    def item_name(self):
        return self.lambda_invocation_name

    def params_for_results_display(self):
        return {
            "puppet_account_id": self.puppet_account_id,
            "lambda_invocation_name": self.lambda_invocation_name,
            "region": self.region,
            "account_id": self.account_id,
        }

    def run(self):
        home_region = config.get_home_region(self.puppet_account_id)
        with self.hub_regional_client(
            "lambda", region_name=home_region
        ) as lambda_client:
            payload = dict(
                account_id=self.account_id,
                region=self.region,
                parameters=self.get_parameter_values(),
            )
            response = lambda_client.invoke(
                FunctionName=self.function_name,
                InvocationType=self.invocation_type,
                Payload=json.dumps(payload),
                Qualifier=self.qualifier,
            )
        success_results = dict(RequestResponse=200, Event=202, DryRun=204)

        if success_results.get(self.invocation_type) != response.get("StatusCode"):
            raise Exception(
                f"{self.lambda_invocation_name} failed for {self.account_id}, {self.region}"
            )
        else:
            if response.get("FunctionError"):
                error_payload = response.get("Payload").read()
                raise Exception(error_payload)
            else:
                self.write_empty_output()



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0
import logging
import os
import traceback

from servicecatalog_puppet import constants
from servicecatalog_puppet.waluigi.cache_download_client import (
    get_cache_download_client,
)
from servicecatalog_puppet.waluigi.event_recorder import record_event


logger = logging.getLogger(constants.PUPPET_LOGGER_NAME)


class TaskExecutorMixin:
    def execute(self):
        if self.should_use_caching:
            if self.complete():
                target = self.output_location_non_cached
                bucket = f"sc-puppet-caching-bucket-{self.puppet_account_id}-{constants.HOME_REGION}"
                key = target
                os.makedirs(os.path.dirname(target), exist_ok=True)
                if not os.path.exists(target):
                    with get_cache_download_client() as s3:
                        s3.download_file(Bucket=bucket, Key=key, Filename=target)
                if not os.path.exists(target):
                    raise Exception(f"{target} was not downloaded from the cache")
            else:
                self.run()
                self.execute()
        else:
            if not self.complete():
                self.run()

    def get_processing_time_details(self):
        task_details = dict(**self.param_kwargs)
        task_details.update(self.params_for_results_display())
        return self.__class__.__name__, task_details

    def on_task_failure(self, exception, duration):
        exception_details = {
            "exception_type": type(exception),
            "exception_stack_trace": traceback.format_exception(
                exception,
                value=exception,
                tb=exception.__traceback__,
            ),
            "duration": duration,
        }
        record_event("failure", self, exception_details)

    def on_task_start(self):
        task_name = self.__class__.__name__
        logger.info(f"{task_name}:{self.task_reference} started")
        record_event("start", self)

    def on_task_success(self, duration):
        record_event("success", self, dict(duration=duration))

    def on_task_timeout(self):
        record_event("timeout", self)

    def on_task_process_failure(self, error_msg):
        exception_details = {
            "exception_type": "PROCESS_FAILURE",
            "exception_stack_trace": error_msg,
        }
        record_event("process_failure", self, exception_details)

    def on_task_broken_task(self, exception):
        record_event("broken_task", self, {"exception": exception})



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

from unittest import skip

from servicecatalog_puppet.workflow import tasks_unit_tests_helper


class DoInvokeLambdaTaskTest(tasks_unit_tests_helper.PuppetTaskUnitTest):
    lambda_invocation_name = "lambda_invocation_name"
    region = "region"
    account_id = "account_id"
    function_name = "function_name"
    qualifier = "qualifier"
    invocation_type = "invocation_type"
    manifest_file_path = "manifest_file_path"

    def setUp(self) -> None:
        from servicecatalog_puppet.workflow.lambda_invocations import (
            do_invoke_lambda_task,
        )

        self.module = do_invoke_lambda_task

        self.sut = self.module.DoInvokeLambdaTask(
            **self.get_common_args(),
            lambda_invocation_name=self.lambda_invocation_name,
            region=self.region,
            account_id=self.account_id,
            function_name=self.function_name,
            qualifier=self.qualifier,
            invocation_type=self.invocation_type,
            manifest_file_path=self.manifest_file_path,
        )

        self.wire_up_mocks()

    def test_params_for_results_display(self):
        # setup
        expected_result = {
            "puppet_account_id": self.puppet_account_id,
            "lambda_invocation_name": self.lambda_invocation_name,
            "region": self.region,
            "account_id": self.account_id,
        }

        # exercise
        actual_result = self.sut.params_for_results_display()

        # verify
        self.assertEqual(expected_result, actual_result)

    @skip
    def test_run(self):
        # setup
        # exercise
        actual_result = self.sut.run()

        # verify
        raise NotImplementedError()



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0
import unittest
from copy import deepcopy
from unittest import skip


class TestManifest(unittest.TestCase):
    account_a = {
        "account_id": "012345678910",
        "default_region": "eu-west-1",
        "name": "accounta",
        "expanded_from": "ou-aaaa-aaaaaaaa",
        "organization": "o-aaaaaaaa",
        "regions_enabled": [
            "eu-west-2",
        ],
        "tags": ["group:A"],
    }
    account_b = {
        "account_id": "009876543210",
        "default_region": "us-west-1",
        "expanded_from": "ou-bbbb-bbbbbbbb",
        "organization": "o-bbbbbbbb",
        "name": "accountb",
        "regions_enabled": [
            "us-west-2",
        ],
        "tags": ["group:B"],
    }
    account_c = {
        "account_id": "432100098765",
        "default_region": "ap-west-1",
        "name": "accountc",
        "regions_enabled": [
            "ap-west-2",
        ],
        "tags": ["group:C"],
    }
    accounts = {
        "accounts": [
            account_a,
            account_b,
            account_c,
        ]
    }
    launch_a = {
        "portfolio": "portfolio_a",
        "product": "product_a",
        "execution": "hub",
        "version": "version_a",
        "deploy_to": {"tags": [{"regions": "default_region", "tag": "group:A"}]},
    }
    assertion_a = {
        "expected": {"foo": "bar"},
        "actual": {"bar": "foo"},
        "assert_for": {"tags": [{"regions": "default_region", "tag": "group:A"}]},
    }
    spoke_local_portfolio_a = {
        "portfolio": "portfolio_a",
        "deploy_to": {"tags": [{"regions": "default_region", "tag": "group:A"}]},
    }
    launch_b = {
        "portfolio": "portfolio_b",
        "product": "product_b",
        "version": "version_b",
        "execution": "hub",
        "deploy_to": {"tags": [{"regions": "default_region", "tag": "group:B"}]},
    }
    launch_c = {
        "portfolio": "portfolio_c",
        "product": "product_c",
        "version": "version_c",
        "execution": "hub",
        "deploy_to": {"tags": [{"regions": "enabled_regions", "tag": "group:B"}]},
    }
    launches = dict(
        launches=dict(launch_a=launch_a, launch_b=launch_b, launch_c=launch_c)
    )
    assertions = dict(assertions=dict(assertion_a=assertion_a))
    spoke_local_portfolios = {
        "spoke-local-portfolios": dict(spoke_local_portfolio_a=spoke_local_portfolio_a)
    }
    puppet_account_id = "0123456789010"

    def setUp(self):
        from servicecatalog_puppet.manifest_utils import Manifest

        self.sut = Manifest()

    def tearDown(self):
        self.sut = None

    def test_get_accounts_by_region(self):
        # setup
        self.sut.update(deepcopy(self.accounts))
        expected_regions = [
            "eu-west-1",
            "eu-west-2",
            "us-west-1",
            "us-west-2",
            "ap-west-1",
            "ap-west-2",
        ]

        # exercise
        actual_result = self.sut.get_accounts_by_region()
        actual_regions = list(actual_result.keys())

        # verify
        self.assertCountEqual(expected_regions, actual_regions)
        self.assertEqual(actual_result.get("eu-west-1"), [self.account_a])
        self.assertIsNone(actual_result.get("eu-west-999"))

    @skip
    def test_get_shares_by_region_portfolio_account(self):
        # setup
        self.sut.update(deepcopy(self.accounts))
        self.sut.update(deepcopy(self.launches))

        # exercise
        actual_result = self.sut.get_shares_by_region_portfolio_account(
            self.puppet_account_id, "launches"
        )

        # verify
        self.assertIsNotNone(actual_result.get("eu-west-1").get("portfolio_a"))
        self.assertEqual(
            actual_result.get("eu-west-1").get("portfolio_a").get("012345678910"),
            self.account_a,
        )
        self.assertIsNotNone(actual_result.get("us-west-1").get("portfolio_b"))
        self.assertEqual(
            actual_result.get("us-west-1").get("portfolio_b").get("009876543210"),
            self.account_b,
        )

    def test_get_sharing_policies_by_region(self):
        # setup
        self.sut.update(deepcopy(self.accounts))
        self.sut.update(deepcopy(self.launches))

        # exercise
        actual_result = self.sut.get_sharing_policies_by_region()

        # verify
        self.assertIsNotNone(actual_result.get("eu-west-1"))
        self.assertEqual(
            actual_result.get("eu-west-1"),
            dict(organizations=["o-aaaaaaaa"], accounts=[]),
        )
        self.assertEqual(
            actual_result.get("us-west-1"),
            dict(organizations=["o-bbbbbbbb"], accounts=[]),
        )
        self.assertEqual(
            actual_result.get("ap-west-1"),
            dict(organizations=[], accounts=["432100098765"]),
        )



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0
import copy
import os

from servicecatalog_puppet import constants


def handle_organizational_units(
    all_tasks,
    all_tasks_task_reference,
    item_name,
    puppet_account_id,
    section_name,
    task_reference,
    task_to_add,
):
    if not task_to_add.get("parent_ou_id"):
        # parent may not exist as it wasn't specified
        parent_path = os.path.dirname(task_to_add.get("path"))
        if parent_path != task_to_add.get("path"):
            parent_task_reference = f"{constants.ORGANIZATIONAL_UNITS}_{parent_path.replace('/', '%2F')}_{task_to_add.get('account_id')}_{task_to_add.get('region')}"
            if not all_tasks.get(parent_task_reference):
                parent_task_to_add = copy.deepcopy(task_to_add)
                parent_task_to_add["task_reference"] = parent_task_reference
                parent_task_to_add["path"] = parent_path
                if parent_path == "/":
                    parent_task_to_add["name"] = parent_path
                else:
                    parent_task_to_add["name"] = os.path.basename(parent_path)
                all_tasks[parent_task_reference] = parent_task_to_add
                handle_organizational_units(
                    all_tasks,
                    all_tasks_task_reference,
                    item_name,
                    puppet_account_id,
                    section_name,
                    parent_task_reference,
                    parent_task_to_add,
                )
            task_to_add["parent_ou_task_ref"] = parent_task_reference
            task_to_add["dependencies_by_reference"].append(parent_task_reference)
